import json
import pathlib


def convert_notebook_to_md(filename, directive='{pyodide}'):
    with open(filename, encoding='utf-8') as nb:
        nb_json = json.loads(nb.read())

    md = f"# {filename.name[:-6].replace('_', ' ').title()}"
    for cell in nb_json['cells']:
        ctype = cell['cell_type']
        if ctype == 'raw':
            continue
        if md:
            md += '\n\n'
        source = cell['source']
        if ctype == 'code':
            md += f'```{directive}\n'
        for src in source:
            md += src
        if ctype == 'code':
            md += '\n```'
    return md

if __name__ == '__main__':
    ROOT = pathlib.Path(__file__).parents[2]
    GALLERY = ROOT / 'doc' / 'gallery'
    for nb in ROOT.glob('./examples/gallery/*.ipynb'):
        md = convert_notebook_to_md(nb)
        with open(GALLERY / f'{nb.name[:-6]}.md', 'w', encoding='utf-8') as f:
            f.write(md)



from __future__ import annotations

import json
import os
import sys
import typing as t

from pathlib import Path

from hatchling.builders.hooks.plugin.interface import BuildHookInterface

BASE_DIR = Path(__file__).parent
GREEN, RED, RESET = "\033[0;32m", "\033[0;31m", "\033[0m"


def build_models():
    from bokeh.ext import build

    print(f"{GREEN}[PANEL]{RESET} Starting building custom models", flush=True)
    panel_dir = BASE_DIR / "panel"
    success = build(panel_dir)
    if sys.platform != "win32":
        # npm can cause non-blocking stdout; so reset it just in case
        import fcntl

        flags = fcntl.fcntl(sys.stdout, fcntl.F_GETFL)
        fcntl.fcntl(sys.stdout, fcntl.F_SETFL, flags & ~os.O_NONBLOCK)

    if success:
        print(f"{GREEN}[PANEL]{RESET} Finished building custom models", flush=True)
    else:
        print(f"{RED}[PANEL]{RESET} Failed building custom models", flush=True)
        sys.exit(1)

def bundle_resources():
    sys.path.insert(0, str(BASE_DIR))
    from panel.compiler import bundle_resources

    print(f"{GREEN}[PANEL]{RESET} Starting bundling custom model resources", flush=True)
    try:
        bundle_resources()
        print(f"{GREEN}[PANEL]{RESET} Finished bundling custom model resources", flush=True)
    except Exception as e:
        print(f"{GREEN}[PANEL]{RESET} Failed bundling custom model resources", flush=True)
        raise e

def clean_js_version(version):
    version = version.replace("-", "")
    for dev in ("a", "b", "rc"):
        version = version.replace(dev + ".", dev)
    return version


def validate_js_version(version):
    # TODO: Double check the logic in this function
    version = version.split(".post")[0]
    with open("./panel/package.json") as f:
        package_json = json.load(f)
    js_version = package_json["version"]
    version = version.split("+")[0]
    if any(dev in version for dev in ("a", "b", "rc")) and "-" not in js_version:
        raise ValueError(f"panel.js dev versions ({js_version}) must separate dev suffix with a dash, e.g. v1.0.0rc1 should be v1.0.0-rc.1.")
    if version != "None" and version != clean_js_version(js_version):
        raise ValueError(f"panel.js version ({js_version}) does not match panel version ({version}). Cannot build release.")


class BuildHook(BuildHookInterface):
    """The hatch build hook."""

    PLUGIN_NAME = "install"

    def initialize(self, version: str, build_data: dict[str, t.Any]) -> None:
        """Initialize the plugin."""
        if self.target_name not in ["wheel", "sdist"]:
            return

        validate_js_version(self.metadata.version)

        if "PANEL_LITE" not in os.environ:
            build_models()
            bundle_resources()



"""
Script to manually test the panels custom bokeh models.

- Models are those defined in index.ts

- Optional installs:
    - ipywidgets
    - ipywidgets_bokeh
    - plotly
    - altair

"""

from io import StringIO

import numpy as np
import param

from bokeh.sampledata.autompg import autompg

import panel as pn

_before = list(locals())

# Model: ace
ace = pn.widgets.Ace(value="import sys", language="python", height=100)

# Model: audio
audio = pn.pane.Audio("http://ccrma.stanford.edu/~jos/mp3/pno-cs.mp3", name="Audio")

# Model: card
_w1 = pn.widgets.TextInput(name="Text:")
_w2 = pn.widgets.FloatSlider(name="Slider")
card = pn.Card(_w1, _w2)

# Model: comm_manager
# Model: customselect

# Model: datetime_picker
datetime_picker = pn.widgets.DatetimePicker()
datetime_range_picker = pn.widgets.DatetimeRangePicker()

# Model: deckgl
_MAPBOX_KEY = "pk.eyJ1IjoicGFuZWxvcmciLCJhIjoiY2s1enA3ejhyMWhmZjNobjM1NXhtbWRrMyJ9.B_frQsAVepGIe-HiOJeqvQ"
_json_spec = {
    "initialViewState": {
        "bearing": -27.36,
        "latitude": 52.2323,
        "longitude": -1.415,
        "maxZoom": 15,
        "minZoom": 5,
        "pitch": 40.5,
        "zoom": 6,
    },
    "layers": [
        {
            "@@type": "HexagonLayer",
            "autoHighlight": True,
            "coverage": 1,
            "data": "https://raw.githubusercontent.com/uber-common/deck.gl-data/master/examples/3d-heatmap/heatmap-data.csv",
            "elevationRange": [0, 3000],
            "elevationScale": 50,
            "extruded": True,
            "getPosition": "@@=[lng, lat]",
            "id": "8a553b25-ef3a-489c-bbe2-e102d18a3211",
            "pickable": True,
        }
    ],
    "mapStyle": "mapbox://styles/mapbox/dark-v9",
    "views": [{"@@type": "MapView", "controller": True}],
}

deck_gl = pn.pane.DeckGL(_json_spec, mapbox_api_key=_MAPBOX_KEY)

# Model: echarts
_echart = {
    "title": {"text": "ECharts entry example"},
    "tooltip": {},
    "legend": {"data": ["Sales"]},
    "xAxis": {"data": ["shirt", "cardign", "chiffon shirt", "pants", "heels", "socks"]},
    "yAxis": {},
    "series": [{"name": "Sales", "type": "bar", "data": [5, 20, 36, 10, 10, 20]}],
}
echart_pane = pn.pane.ECharts(_echart, height=480, width=640)

# Model: file_downloadm
_sio = StringIO()
autompg.to_csv(_sio)
_sio.seek(0)
file_download = pn.widgets.FileDownload(_sio, embed=True, filename="autompg.csv")

# Model: html (py: Markup)
html = pn.pane.HTML("<h1>TEST<h1>")

# Model: ipywidget
try:
    import ipywidgets as _ipw
    import ipywidgets_bokeh as _ipwb  # noqa

    ipywidget = pn.pane.IPyWidget(_ipw.FloatSlider(description="Float"))
except ImportError:
    ipywidget = "Need to have ipywidgets and ipywidgets_bokeh installed"

# Model: json (py: Markup)
json = pn.pane.JSON({"test": 1, "B": ["1", None]})

# Model: json_editor
json_editor = pn.widgets.JSONEditor(
    value={
        "dict": {"key": "value"},
        "float": 3.14,
        "int": 1,
        "list": [1, 2, 3],
        "string": "A string",
    },
    width=500,
)

# Model: katex
latex1 = pn.pane.LaTeX(
    "The LaTeX pane supports two delimiters: $LaTeX$ and \(LaTeX\)",
    styles={"font-size": "18pt"},
    width=800,
)

# Model: location

# Model: mathjax
latex2 = pn.pane.LaTeX(
    "$\sum_{j}{\sum_{i}{a*w_{j, i}}}$", renderer="mathjax", styles={"font-size": "18pt"}
)

# Model: perspective
_data = {"x": [1, 2, 3], "y": [1, 2, 3]}
perspective = pn.pane.Perspective(_data)

# Model: player
player = pn.widgets.Player(
    name="Player", start=0, end=100, value=32, loop_policy="loop"
)

# Model: plotly
try:
    import plotly.express as _px

    plotly = pn.pane.Plotly(
        _px.line({"Day": range(7), "Orders": range(7)}, x="Day", y="Orders")
    )
except ImportError:
    plotly = "Need to have plotly installed"

# Model: progress
progress = pn.indicators.Progress(name="Progress", value=20, width=200, height=20)

# Model: quill
text_editor = pn.widgets.TextEditor(placeholder="Enter some text", width=500)

# Model: reactive_html
class _Slideshow(pn.reactive.ReactiveHTML):

    index = param.Integer(default=0)

    _template = '<img id="slideshow" src="https://picsum.photos/800/300?image=${index}" onclick="${_img_click}"></img>'

    def _img_click(self, event):
        self.index += 1


_Slideshow.name = "ReactiveHTML <br> (not working in html)"
reactive = _Slideshow(width=800, height=300)

# Model: singleselect
single_select = pn.widgets.Select(value="A", options=list("ABCD"))

# Model: speech_to_text
# speech_to_text_basic = pn.widgets.SpeechToText(button_type="light")

# Model: state
# Model: tabs
_w1 = pn.widgets.TextInput(name="Text:")
_w2 = pn.widgets.FloatSlider(name="Slider")
tabs = pn.Tabs(_w1, _w2)

# Model: tabulator
tabulator = pn.widgets.Tabulator(autompg)


# Model: terminal
terminal = pn.widgets.Terminal(
    "Welcome to the Panel Terminal!\nI'm based on xterm.js\n\n",
    options={"cursorBlink": True},
    height=300,
    sizing_mode="stretch_width",
)

# Model: text_to_speech
# text_to_speech = pn.widgets.TextToSpeech(name="Speech Synthesis")

# Model: trend
_data = {"x": np.arange(50), "y": np.random.randn(50).cumsum()}
trend = pn.indicators.Trend(name="Price", data=_data, width=200, height=200)

# Model: vega
try:
    import altair as _alt

    _chart = (
        _alt.Chart(autompg)
        .mark_circle(size=60)
        .encode(
            x="hp", y="mpg", color="origin", tooltip=["name", "origin", "hp", "mpg"]
        )
        .interactive()
    )

    vega = pn.pane.Vega(_chart)
except ImportError:
    vega = "Need to have altair installed"


# Model: video
video = pn.pane.Video(
    "https://file-examples.com/storage/fe2333f3be630e8e7965da7/2017/04/file_example_MP4_480_1_5MG.mp4",
    width=640,
    height=360,
    loop=True,
)

# Model: videostream
video_stream = pn.widgets.VideoStream(name="Video Stream")

# Model: vtk.vtkjs
# Model: vtk.vtkvolume
# Model: vtk.vtkaxes
# Model: vtk.vtksynchronized


# Combine and save to html
widgets = [v for k, v in locals().items() if k not in _before and not k.startswith("_")]
names = [getattr(w.__class__, "name", w) for w in widgets]
combined = pn.Column(
    *[
        pn.Column(
            pn.Row(pn.Column(n, width=300), w),
            pn.layout.Divider(),
            sizing_mode="stretch_width",
        )
        for w, n, in zip(widgets, names)
    ]
)

if __name__ == "__main__":
    from bokeh.resources import INLINE

    combined.save("models.html", resources=INLINE)
elif __name__.startswith("bokeh"):
    combined.servable()



from __future__ import annotations

from typing import Any

import bokeh.core.properties as bp

from bokeh.events import ModelEvent
from bokeh.model import DataModel

from ..config import config
from ..io.resources import bundled_files
from ..util import classproperty
from .layout import HTMLBox


class DataEvent(ModelEvent):

    event_name = 'data_event'

    def __init__(self, model, data=None):
        self.data = data
        super().__init__(model=model)

    def event_values(self) -> dict[str, Any]:
        return dict(super().event_values(), data=self.data)


class ESMEvent(DataEvent):

    event_name = 'esm_event'


class ReactiveESM(HTMLBox):

    bundle = bp.Nullable(bp.String)

    class_name = bp.String()

    children = bp.List(bp.String)

    data = bp.Instance(DataModel)

    dev = bp.Bool(False)

    esm = bp.String()

    importmap = bp.Dict(bp.String, bp.Dict(bp.String, bp.String))

    __javascript_raw__ = [
        f"{config.npm_cdn}/es-module-shims@^1.10.0/dist/es-module-shims.min.js"
    ]

    @classproperty
    def __javascript__(cls):
        return bundled_files(cls)


class ReactComponent(ReactiveESM):
    """
    Renders jsx/tsx based ESM bundles using React.
    """


class AnyWidgetComponent(ReactComponent):
    """
    Renders AnyWidget esm definitions by adding a compatibility layer.
    """



import pathlib

from base64 import b64encode
from io import BytesIO

import numpy as np
import pytest

try:
    from scipy.io import wavfile
except Exception:
    wavfile = None

from panel.pane import Audio, Video
from panel.pane.media import TensorLike, _is_1dim_int_or_float_tensor

ASSETS = pathlib.Path(__file__).parent / 'assets'

scipy_available = pytest.mark.skipif(wavfile is None, reason="requires scipy")



def test_video_url(document, comm):
    url = 'https://file-examples-com.github.io/uploads/2017/04/file_example_MP4_640_3MG.mp4'
    video = Video(url)
    model = video.get_root(document, comm=comm)

    # To check if url is send to the bokeh model
    assert model.value == url


def test_local_video(document, comm):
    video = Video(str(ASSETS / 'mp4.mp4'))
    model = video.get_root(document, comm=comm)

    assert model.value == 'data:video/mp4;base64,AAAAIGZ0eXBpc29tAAACAGlzb21pc28yYXZjMW1wNDEAAAAIZnJlZQAAAAhtZGF0AAAA1m1vb3YAAABsbXZoZAAAAAAAAAAAAAAAAAAAA+gAAAAAAAEAAAEAAAAAAAAAAAAAAAABAAAAAAAAAAAAAAAAAAAAAQAAAAAAAAAAAAAAAAAAQAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAIAAABidWR0YQAAAFptZXRhAAAAAAAAACFoZGxyAAAAAAAAAABtZGlyYXBwbAAAAAAAAAAAAAAAAC1pbHN0AAAAJal0b28AAAAdZGF0YQAAAAEAAAAATGF2ZjU3LjQxLjEwMA==' # noqa


def test_video_autoplay(document, comm):
    video = Video(str(ASSETS / 'mp4.mp4'), autoplay=True)
    model = video.get_root(document, comm=comm)

    assert model.autoplay

def test_video_muted(document, comm):
    video = Video(str(ASSETS / 'mp4.mp4'), muted=True)
    model = video.get_root(document, comm=comm)

    assert model.muted


def test_local_audio(document, comm):
    audio = Audio(str(ASSETS / 'mp3.mp3'))
    model = audio.get_root(document, comm=comm)

    assert model.value == 'data:audio/mp3;base64,/+MYxAAAAANIAAAAAExBTUUzLjk4LjIAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA' # noqa


@scipy_available
def test_numpy_audio(document, comm):
    sps = 8000 # Samples per second
    duration = 0.01 # Duration in seconds

    modulator_frequency = 2.0
    carrier_frequency = 120.0
    modulation_index = 2.0

    time = np.arange(sps*duration) / sps
    modulator = np.sin(2.0 * np.pi * modulator_frequency * time) * modulation_index
    waveform = np.sin(2. * np.pi * (carrier_frequency * time + modulator))

    waveform_quiet = waveform * 0.3
    waveform_int = np.int16(waveform_quiet * 32767)

    audio = Audio(waveform_int, sample_rate=sps)

    model = audio.get_root(document, comm=comm)

    assert model.value == 'data:audio/wav;base64,UklGRsQAAABXQVZFZm10IBAAAAABAAEAQB8AAIA+AAACABAAZGF0YaAAAAAAAF4ErQjgDOgQuBRDGH0bXB7WIOMifCScJT8mYyYHJi0l1yMLIs0fJR0dGr4WFBMqDw4LzQZ1Ahf+vfl59VfxZu2z6UrmN+OD4DjeXNz42g7aotm22UjaWNvi3ODeTOEe5E3nzuqU7pXywvYO+2v/yAMZCFAMXhA1FMoXDxv7HYMgoCJJJHolLyZlJhwmVSURJFciKiCTHZoaSBeqE8oP' # noqa


@scipy_available
def test_audio_array(document, comm):
    data = np.random.randint(-100,100, 100).astype('int16')
    sample_rate = 10
    buffer = BytesIO()
    wavfile.write(buffer, sample_rate, data)
    b64_encoded = b64encode(buffer.getvalue()).decode('utf-8')

    audio = Audio(data, sample_rate=sample_rate)
    model = audio.get_root(document, comm=comm)
    model_value = model.value

    assert model_value.split(',')[1] == b64_encoded
    assert model.value.startswith('data:audio/wav;base64')


def test_audio_url(document, comm):
    audio_url = 'http://ccrma.stanford.edu/~jos/mp3/pno-cs.mp3'
    audio = Audio(audio_url)
    model = audio.get_root(document, comm=comm)

    assert audio_url == model.value

def test_audio_muted(document, comm):
    audio_url = 'http://ccrma.stanford.edu/~jos/mp3/pno-cs.mp3'
    audio = Audio(audio_url, muted=True)
    model = audio.get_root(document, comm=comm)

    assert model.muted

def test_audio_autoplay(document, comm):
    audio_url = 'http://ccrma.stanford.edu/~jos/mp3/pno-cs.mp3'
    audio = Audio(audio_url, autoplay=True)
    model = audio.get_root(document, comm=comm)

    assert model.autoplay

def get_audio_np_float64(duration=0.01):
    sample_rate = Audio.sample_rate
    time_variable = np.linspace(0, duration, int(duration * sample_rate), endpoint=False)
    sine_wave_400hz = 0.5 * np.sin(2 * np.pi * 440 * time_variable)
    return sine_wave_400hz

def get_audio_np_float32(duration=0.01):
    return get_audio_np_float64(duration=duration).astype(np.float32)

@scipy_available
def test_audio_float64(document, comm):
    audio = Audio(object=get_audio_np_float64())
    model = audio.get_root(document, comm=comm)
    assert model.value == 'data:audio/wav;base64,UklGRpYDAABXQVZFZm10IBAAAAABAAEARKwAAIhYAQACABAAZGF0YXIDAAAAAAIEAAj3C+EPuxOCFzEbxB45Iosltyi7K5IuOzGyM/U1AjjWOXE7zzzxPdQ+dz/cP/8/4z+GP+k+DT7zPJs7CDo6ODM29jOFMeIuECwRKeklmyIqH5ob7hcqFFIQaQx0CHYEdABz/HT4fPSQ8LTs6+g55aLhKt7U2qPXm9S+0RDPk8xKyjfIXMa7xFbDLcJDwZjALcABwBfAbMACwdbB6sI6xMfFj8ePycbLMc7P0JzTlta52QPdcOD946bnZ+s97yTzGPcV+xf/GQMZBxEL/g7dEqgWXRr3HXMhzSQCKA8r8S2kMCczdjWPN3E5GTuFPLQ9pj5YP8o//T/vP6E/Ez9FPjk97ztpOqg4rjZ9NBgygC+4LMMppCZfI/UfbBzGGAcVMxFODVsJXwVeAVz9XPli9XPxk+3F6Q7mceLx3pPbWdhI1WHSqM8gzcvKq8jDxhXFosNrwnPBucA/wAbADMBTwNrAoMGmwunDaMUixxbJQMugzTPQ9dLl1f/YQNym3yzjzuaL6lzuQPIx9i36Lv4wAjAGKwobDv0RzRWHGSgdqyANJEsnYSpNLQswmTL0NBo3CDm9Ojc8dT10PjU/tj/3P/g/uD85P3k+ez0/PMc6EzkmNwE1qDIbMF4tcypdJyAkvyA8HZ0Z4xUUEjIOQgpIBkcCRf5E+kj2V/Jz7qHq5OZA47rfVNwR2fbVBtNC0K/NTssiyS3HccXxw6zCpsHewFXADcAFwD3AtsBuwWXCmsMMxbnGn8i+yhHNmc9Q0jbVR9iA293eXOL55a/pfe1c8Uv1RPlE/UYBSAVECTcNHRHxFLEYVxzhH0sjkiaxKacscC8JMnA0ojadOF865zsyPUA+Dz+eP+4//T/MP1s/qj66PYw8ITt7OZs3gzU1M7MwAS4gKxQo4CSGIQsechq+FvMSFQ8oCzAHMAMu/yz7L/c781Tvfeu75xLkhOAW3czZqNat09/QQM7Ty5vJmsfRxUPE8cLcwQbBb8AYwAHAK8CVwD7BJ8JOw7LEUsYsyD7KhcwBz67RitSR18HaFt6O4STl1eie7HrwZfRc+Fv8XQBfBF0IUgw7EBQU2ReFGxYfhyLWJf8o/yvSLnYx6TMnNi84/jmTO+w8CD7lPoM/4T//P90/ej/YPvY91jx5O+A5DTgBNsAzSjGiLswrySg=' # noqa

@scipy_available
def test_audio_float32(document, comm):
    audio = Audio(object=get_audio_np_float32())
    model = audio.get_root(document, comm=comm)
    assert model.value == 'data:audio/wav;base64,UklGRpYDAABXQVZFZm10IBAAAAABAAEARKwAAIhYAQACABAAZGF0YXIDAAAAAAIEAAj3C+EPuxOCFzEbxB45Iosltyi7K5IuOzGyM/U1AjjWOXE7zzzxPdQ+dz/cP/8/4z+GP+k+DT7zPJs7CDo6ODM29jOFMeIuECwRKeklmyIqH5ob7hcqFFIQaQx0CHYEdABz/HT4fPSQ8LTs6+g55aLhKt7U2qPXm9S+0RDPk8xKyjfIXMa7xFbDLcJDwZjALcABwBfAbMACwdbB6sI6xMfFj8ePycbLMc7P0JzTlta52QPdcOD946bnZ+s97yTzGPcV+xf/GQMZBxEL/g7dEqgWXRr3HXMhzSQCKA8r8S2kMCczdjWPN3E5GTuFPLQ9pj5YP8o//T/vP6E/Ez9FPjk97ztpOqg4rjZ9NBgygC+4LMMppCZfI/UfbBzGGAcVMxFODVsJXwVeAVz9XPli9XPxk+3F6Q7mceLx3pPbWdhI1WHSqM8gzcvKq8jDxhXFosNrwnPBucA/wAbADMBTwNrAoMGmwunDaMUixxbJQMugzTPQ9dLl1f/YQNym3yzjzuaL6lzuQPIx9i36Lv4wAjAGKwobDv0RzRWHGSgdqyANJEsnYSpNLQswmTL0NBo3CDm9Ojc8dT10PjU/tj/3P/g/uD85P3k+ez0/PMc6EzkmNwE1qDIbMF4tcypdJyAkvyA8HZ0Z4xUUEjIOQgpIBkcCRf5E+kj2V/Jz7qHq5OZA47rfVNwR2fbVBtNC0K/NTssiyS3HccXxw6zCpsHewFXADcAFwD3AtsBuwWXCmsMMxbnGn8i+yhHNmc9Q0jbVR9iA293eXOL55a/pfe1c8Uv1RPlE/UYBSAVECTcNHRHxFLEYVxzhH0sjkiaxKacscC8JMnA0ojadOF865zsyPUA+Dz+eP+4//T/MP1s/qj66PYw8ITt7OZs3gzU1M7MwAS4gKxQo4CSGIQsechq+FvMSFQ8oCzAHMAMu/yz7L/c781Tvfeu75xLkhOAW3czZqNat09/QQM7Ty5vJmsfRxUPE8cLcwQbBb8AYwAHAK8CVwD7BJ8JOw7LEUsYsyD7KhcwBz67RitSR18HaFt6O4STl1eie7HrwZfRc+Fv8XQBfBF0IUgw7EBQU2ReFGxYfhyLWJf8o/yvSLnYx6TMnNi84/jmTO+w8CD7lPoM/4T//P90/ej/YPvY91jx5O+A5DTgBNsAzSjGiLswrySg='  # noqa


class TensorMock:
    def __init__(self, duration=2.0):
        self._data = get_audio_np_float32(duration=duration)

    def numpy(self):
        return self._data

    def dim(self):
        return self._data.ndim

    @property
    def dtype(self):
        return "torch.float32"

def test_torch_like():
    mock = TensorMock()

    assert not isinstance(None, TensorLike)
    assert isinstance(mock, TensorLike)

def get_audio_tensor_float32(duration=2.0):
    return TensorMock(duration=duration)

@scipy_available
def test_audio_tensor_float32(document, comm):
    obj = get_audio_tensor_float32(duration=0.01)
    assert _is_1dim_int_or_float_tensor(obj)
    audio = Audio(object=obj)
    model = audio.get_root(document, comm=comm)
    assert model.value == 'data:audio/wav;base64,UklGRpYDAABXQVZFZm10IBAAAAABAAEARKwAAIhYAQACABAAZGF0YXIDAAAAAAIEAAj3C+EPuxOCFzEbxB45Iosltyi7K5IuOzGyM/U1AjjWOXE7zzzxPdQ+dz/cP/8/4z+GP+k+DT7zPJs7CDo6ODM29jOFMeIuECwRKeklmyIqH5ob7hcqFFIQaQx0CHYEdABz/HT4fPSQ8LTs6+g55aLhKt7U2qPXm9S+0RDPk8xKyjfIXMa7xFbDLcJDwZjALcABwBfAbMACwdbB6sI6xMfFj8ePycbLMc7P0JzTlta52QPdcOD946bnZ+s97yTzGPcV+xf/GQMZBxEL/g7dEqgWXRr3HXMhzSQCKA8r8S2kMCczdjWPN3E5GTuFPLQ9pj5YP8o//T/vP6E/Ez9FPjk97ztpOqg4rjZ9NBgygC+4LMMppCZfI/UfbBzGGAcVMxFODVsJXwVeAVz9XPli9XPxk+3F6Q7mceLx3pPbWdhI1WHSqM8gzcvKq8jDxhXFosNrwnPBucA/wAbADMBTwNrAoMGmwunDaMUixxbJQMugzTPQ9dLl1f/YQNym3yzjzuaL6lzuQPIx9i36Lv4wAjAGKwobDv0RzRWHGSgdqyANJEsnYSpNLQswmTL0NBo3CDm9Ojc8dT10PjU/tj/3P/g/uD85P3k+ez0/PMc6EzkmNwE1qDIbMF4tcypdJyAkvyA8HZ0Z4xUUEjIOQgpIBkcCRf5E+kj2V/Jz7qHq5OZA47rfVNwR2fbVBtNC0K/NTssiyS3HccXxw6zCpsHewFXADcAFwD3AtsBuwWXCmsMMxbnGn8i+yhHNmc9Q0jbVR9iA293eXOL55a/pfe1c8Uv1RPlE/UYBSAVECTcNHRHxFLEYVxzhH0sjkiaxKacscC8JMnA0ojadOF865zsyPUA+Dz+eP+4//T/MP1s/qj66PYw8ITt7OZs3gzU1M7MwAS4gKxQo4CSGIQsechq+FvMSFQ8oCzAHMAMu/yz7L/c781Tvfeu75xLkhOAW3czZqNat09/QQM7Ty5vJmsfRxUPE8cLcwQbBb8AYwAHAK8CVwD7BJ8JOw7LEUsYsyD7KhcwBz67RitSR18HaFt6O4STl1eie7HrwZfRc+Fv8XQBfBF0IUgw7EBQU2ReFGxYfhyLWJf8o/yvSLnYx6TMnNi84/jmTO+w8CD7lPoM/4T//P90/ej/YPvY91jx5O+A5DTgBNsAzSjGiLswrySg='  # noqa



from panel.custom import Child, ReactComponent


class Example(ReactComponent):

    child = Child()

    _esm = """
    export function render({ model }) {
      return <button>{model.get_child("child")}</button>
    }"""

Example(child='A Markdown pane!').servable()



import pytest


@pytest.fixture
def mock_completion() -> dict:
    return {
        "id": "chatcmpl-7fcZavknQda3SQ",
        "object": "chat.completion",
        "created": 1689989000,
        "model": "gpt-3.5-turbo-0613",
        "choices": [
            {
                "index": 0,
                "message": {
                    "role": "assistant",
                    "content": "Bar Baz",
                },
                "finish_reason": "stop",
            }
        ],
    }



from panel.chat.icon import ChatReactionIcons


class TestChatReactionIcons:
    def test_init_default(self):
        icons = ChatReactionIcons(width=50, height=50)
        assert icons.options == {"favorite": "heart"}
        assert "favorite" in icons._rendered_icons
        assert icons._rendered_icons["favorite"].icon == "heart"
        assert icons._rendered_icons["favorite"].active_icon == ""
        assert len(icons._composite) == 1

    def test_options(self):
        icons = ChatReactionIcons(options={"like": "thumb-up", "dislike": "thumb-down"})
        assert icons.options == {"like": "thumb-up", "dislike": "thumb-down"}
        assert "like" in icons._rendered_icons
        assert icons._rendered_icons["like"].icon == "thumb-up"
        assert icons._rendered_icons["like"].active_icon == ""
        assert "dislike" in icons._rendered_icons
        assert icons._rendered_icons["dislike"].icon == "thumb-down"
        assert icons._rendered_icons["dislike"].active_icon == ""
        assert len(icons._composite) == 1
        assert len(icons._composite[0]) == 2

        icons.options = {"favorite": "heart"}
        assert icons.options == {"favorite": "heart"}
        assert "favorite" in icons._rendered_icons
        assert icons._rendered_icons["favorite"].icon == "heart"
        assert icons._rendered_icons["favorite"].active_icon == ""
        assert len(icons._composite) == 1
        assert len(icons._composite[0]) == 1

    def test_value(self):
        icons = ChatReactionIcons(
            options={"like": "thumb-up", "dislike": "thumb-down"}, value=["like"]
        )
        assert icons.value == ["like"]
        assert icons._rendered_icons["like"].value is True
        assert icons._rendered_icons["dislike"].value is False

        icons.value = ["dislike"]
        assert icons.value == ["dislike"]
        assert icons._rendered_icons["like"].value is False
        assert icons._rendered_icons["dislike"].value is True

    def test_active_icons(self):
        icons = ChatReactionIcons(
            options={"like": "thumb-up"},
            active_icons={"like": "thumb-down"},
            value=["like"],
        )
        assert icons.active_icons == {"like": "thumb-down"}
        assert icons._rendered_icons["like"].active_icon == "thumb-down"
        assert len(icons._composite) == 1

        icons.active_icons = {"like": "heart"}
        assert icons.active_icons == {"like": "heart"}
        assert icons._rendered_icons["like"].active_icon == "heart"
        assert len(icons._composite) == 1



import pytest

pytest.importorskip("playwright")


from panel.chat import ChatReactionIcons
from panel.tests.util import serve_component, wait_until

pytestmark = pytest.mark.ui


def test_toggle_icon_click(page):
    icons = ChatReactionIcons(options={"like": "thumb-up", "dislike": "thumb-down"}, value=["dislike"])
    serve_component(page, icons)

    assert icons.value == ["dislike"]
    page.locator(".ti-thumb-up").click()
    wait_until(lambda: icons.value == ["dislike", "like"], page)

    page.locator(".ti-thumb-down-filled").click()
    wait_until(lambda: icons.value == ["like"], page)

    page.locator(".ti-thumb-up-filled").click()
    wait_until(lambda: icons.value == [], page)

    page.locator(".ti-thumb-up").click()
    wait_until(lambda: icons.value == ["like"], page)



from panel.custom import Children, JSComponent


class Example(JSComponent):

    children = Children()

    _esm = """
    export function render({ model }) {
      const div = document.createElement('div')
      div.append(...model.get_child('children'))
      return div
    }"""

Example(children=[
    'A Markdown pane!',
    'Another Markdown pane!'
]).servable()



import json
import pathlib
import subprocess

package_json = json.loads(
    (pathlib.Path(__file__).parent.parent / 'panel' / 'package.json').read_text()
)
js_version = package_json['version']

sp = subprocess.Popen(['aws', 's3', 'sync', 'panel/dist', f's3://cdn.holoviz.org/panel/{js_version}/dist/'])
sp.wait()
sp2 = subprocess.Popen(['aws', 's3', 'cp', 'panel/dist/wheels/', 's3://cdn.holoviz.org/panel/wheels/', '--recursive', '--exclude', '"*"', '--include', '"bokeh*"'])
sp2.wait()
sp3 = subprocess.Popen(['aws', 's3', 'cp', 'panel/package.json', f's3://cdn.holoviz.org/panel/{js_version}/package.json'])
sp3.wait()



import asyncio

from io import BytesIO

import pytest
import requests

from panel.chat.input import ChatAreaInput
from panel.chat.interface import ChatInterface
from panel.chat.message import ChatMessage
from panel.layout import Row, Tabs
from panel.pane import Image
from panel.tests.util import async_wait_until, wait_until
from panel.widgets.button import Button
from panel.widgets.input import FileInput, TextAreaInput, TextInput
from panel.widgets.select import RadioButtonGroup

ChatInterface.callback_exception = "raise"


class TestChatInterface:
    @pytest.fixture
    def chat_interface(self):
        return ChatInterface()

    def test_init(self, chat_interface):
        assert len(chat_interface._button_data) == 5
        assert len(chat_interface._widgets) == 1
        assert isinstance(chat_interface._input_layout, Row)
        assert isinstance(chat_interface._widgets["ChatAreaInput"], ChatAreaInput)

        assert chat_interface.active == -1

        # Buttons added to input layout
        inputs = chat_interface._input_layout
        for index, button_data in enumerate(chat_interface._button_data.values()):
            widget = inputs[index + 1]
            assert isinstance(widget, Button)
            assert widget.name == button_data.name.title()

    def test_init_avatar_image(self, chat_interface):
        chat_interface.avatar = Image("https://panel.holoviz.org/_static/logo_horizontal.png")
        assert chat_interface.avatar.object == "https://panel.holoviz.org/_static/logo_horizontal.png"

    @pytest.mark.internet
    @pytest.mark.parametrize("type_", [bytes, BytesIO])
    def test_init_avatar_bytes(self, type_, chat_interface):
        with requests.get("https://panel.holoviz.org/_static/logo_horizontal.png") as resp:
            chat_interface.avatar = type_(resp.content)
        assert isinstance(chat_interface.avatar, type_)

    def test_init_custom_widgets(self):
        widgets = [TextInput(name="Text"), FileInput()]
        chat_interface = ChatInterface(widgets=widgets)
        assert len(chat_interface._widgets) == 2
        assert isinstance(chat_interface._input_layout, Tabs)
        assert isinstance(chat_interface._widgets["Text"], TextInput)
        assert isinstance(chat_interface._widgets["FileInput"], FileInput)
        assert chat_interface.active == 0

    def test_active_in_constructor(self):
        widgets = [TextInput(name="Text"), FileInput()]
        chat_interface = ChatInterface(widgets=widgets, active=1)
        assert chat_interface.active == 1

    def test_file_input_only(self):
        ChatInterface(widgets=[FileInput(name="CSV File", accept=".csv")])

    def test_active_widget(self, chat_interface):
        active_widget = chat_interface.active_widget
        assert isinstance(active_widget, ChatAreaInput)

    def test_active(self):
        widget = TextInput(name="input")
        chat_interface = ChatInterface(widgets=[widget])
        assert chat_interface.active == -1

    def test_active_multiple_widgets(self, chat_interface):
        widget1 = TextInput(name="input1")
        widget2 = TextInput(name="input2")
        chat_interface.widgets = [widget1, widget2]
        assert chat_interface.active == 0

        chat_interface.active = 1
        assert chat_interface.active == 1
        assert isinstance(chat_interface.active_widget, TextInput)

    def test_click_send(self, chat_interface: ChatInterface):
        chat_interface.widgets = [TextAreaInput()]
        chat_interface.active_widget.value = "Message"
        # since it's TextAreaInput and NOT TextInput, need to manually send
        assert len(chat_interface.objects) == 0
        chat_interface._click_send(None)
        assert len(chat_interface.objects) == 1

    def test_click_send_with_no_value_input(self, chat_interface: ChatInterface):
        chat_interface.widgets = [RadioButtonGroup(options=["A", "B"])]
        chat_interface.active_widget.value = "A"
        chat_interface._click_send(None)
        assert chat_interface.objects[0].object == "A"

    def test_show_stop_disabled(self, chat_interface: ChatInterface):
        async def callback(msg, user, instance):
            yield "A"
            send_button = instance._buttons["send"]
            stop_button = instance._buttons["stop"]
            wait_until(lambda: send_button.visible)
            wait_until(lambda: send_button.disabled) #  should be disabled while callback is running
            assert not stop_button.visible
            yield "B"  # should not stream this

        chat_interface.callback = callback
        chat_interface.show_stop = False
        chat_interface.send("Message", respond=True)
        send_button = chat_interface._input_layout[1]
        stop_button = chat_interface._input_layout[2]
        assert send_button.name == "Send"
        assert stop_button.name == "Stop"
        assert send_button.visible
        assert not send_button.disabled
        assert not stop_button.visible

    def test_show_stop_for_async(self, chat_interface: ChatInterface):
        async def callback(msg, user, instance):
            send_button = instance._buttons["send"]
            stop_button = instance._buttons["stop"]
            await async_wait_until(lambda: stop_button.visible)
            await async_wait_until(lambda: not send_button.visible)

        chat_interface.callback = callback
        chat_interface.send("Message", respond=True)
        send_button = chat_interface._input_layout[1]
        assert not send_button.disabled

    def test_show_stop_for_async_generator(self, chat_interface: ChatInterface):
        async def callback(msg, user, instance):
            send_button = instance._buttons["send"]
            stop_button = instance._buttons["stop"]
            await async_wait_until(lambda: stop_button.visible)
            await async_wait_until(lambda: not send_button.visible)
            yield "Hello"

        chat_interface.callback = callback
        chat_interface.send("Message", respond=True)
        send_button = chat_interface._input_layout[1]
        assert not send_button.disabled

    def test_show_stop_for_sync_generator(self, chat_interface: ChatInterface):
        def callback(msg, user, instance):
            send_button = instance._buttons["send"]
            stop_button = instance._buttons["stop"]
            wait_until(lambda: stop_button.visible)
            wait_until(lambda: not send_button.visible)
            yield "Hello"

        chat_interface.callback = callback
        chat_interface.send("Message", respond=True)
        send_button = chat_interface._input_layout[1]
        assert not send_button.disabled

    def test_click_stop(self, chat_interface: ChatInterface):
        async def callback(msg, user, instance):
            send_button = instance._buttons["send"]
            stop_button = instance._buttons["stop"]
            await async_wait_until(lambda: stop_button.visible)
            await async_wait_until(lambda: not send_button.visible)
            instance._click_stop(None)

        chat_interface.callback = callback
        chat_interface.placeholder_threshold = 0.001
        try:
            chat_interface.send("Message", respond=True)
        except asyncio.exceptions.CancelledError:
            pass
        wait_until(lambda: not chat_interface._buttons["send"].disabled)
        wait_until(lambda: chat_interface._buttons["send"].visible)
        wait_until(lambda: not chat_interface._buttons["stop"].visible)

    @pytest.mark.parametrize("widget", [TextInput(), TextAreaInput()])
    def test_auto_send_types(self, chat_interface: ChatInterface, widget):
        chat_interface.auto_send_types = [TextAreaInput]
        chat_interface.widgets = [widget]
        chat_interface.active_widget.value = "Message"
        assert len(chat_interface.objects) == 1
        assert chat_interface.objects[0].object == "Message"

    def test_click_undo(self, chat_interface):
        chat_interface.user = "User"
        chat_interface.send("Message 1")
        chat_interface.send("Message 2")
        chat_interface.send("Message 3", user="Assistant")
        expected = chat_interface.objects[-2:].copy()
        chat_interface._click_undo(None)
        assert len(chat_interface.objects) == 1
        assert chat_interface.objects[0].object == "Message 1"
        assert chat_interface._button_data["undo"].objects == expected

        # revert
        chat_interface._click_undo(None)
        assert len(chat_interface.objects) == 3
        assert chat_interface.objects[0].object == "Message 1"
        assert chat_interface.objects[1].object == "Message 2"
        assert chat_interface.objects[2].object == "Message 3"

    def test_click_clear(self, chat_interface):
        chat_interface.send("Message 1")
        chat_interface.send("Message 2")
        chat_interface.send("Message 3")
        expected = chat_interface.objects.copy()
        chat_interface._click_clear(None)
        assert len(chat_interface.objects) == 0
        assert chat_interface._button_data["clear"].objects == expected

    def test_click_rerun(self, chat_interface):
        self.count = 0

        def callback(contents, user, instance):
            self.count += 1
            return self.count

        chat_interface.callback = callback
        chat_interface.send("Message 1")
        wait_until(lambda: len(chat_interface.objects) >= 2)
        wait_until(lambda: chat_interface.objects[1].object == 1)
        chat_interface._click_rerun(None)
        wait_until(lambda: chat_interface.objects[1].object == 2)

    def test_click_rerun_null(self, chat_interface):
        chat_interface._click_rerun(None)
        assert len(chat_interface.objects) == 0

    def test_replace_widgets(self, chat_interface):
        assert isinstance(chat_interface._input_layout, Row)

        chat_interface.widgets = [TextAreaInput(), FileInput()]
        assert len(chat_interface._widgets) == 2
        assert isinstance(chat_interface._input_layout, Tabs)
        assert isinstance(chat_interface._widgets["TextAreaInput"], TextAreaInput)
        assert isinstance(chat_interface._widgets["FileInput"], FileInput)

    def test_reset_on_send(self, chat_interface):
        chat_interface.active_widget.value = "Hello"
        chat_interface.reset_on_send = True
        assert chat_interface.active_widget.value == ""

    def test_reset_on_send_text_area(self, chat_interface):
        chat_interface.widgets = TextAreaInput()
        chat_interface.reset_on_send = False
        chat_interface.active_widget.value = "Hello"
        assert chat_interface.active_widget.value == "Hello"

    def test_widgets_supports_list_and_widget(self, chat_interface):
        chat_interface.widgets = TextAreaInput()
        chat_interface.widgets = [TextAreaInput(), FileInput]

    def test_show_button_name_width(self, chat_interface):
        assert chat_interface.show_button_name
        assert chat_interface.width is None
        chat_interface.width = 200
        assert chat_interface.show_button_name
        assert chat_interface._input_layout[1].name == "Send"

    def test_show_button_name_set(self, chat_interface):
        chat_interface.show_button_name = False
        chat_interface.width = 800
        assert not chat_interface.show_button_name
        assert chat_interface._input_layout[1].name == ""

    def test_show_send_interactive(self, chat_interface):
        send_button = chat_interface._input_layout[1]
        assert chat_interface.show_send
        assert send_button.visible
        chat_interface.show_send = False
        assert not chat_interface.show_send
        assert not send_button.visible

    @pytest.mark.parametrize("key", ["callback", "post_callback"])
    def test_button_properties_new_button(self, chat_interface, key):
        def callback(instance, event):
            instance.send("Checking if this works", respond=False)

        chat_interface.widgets = TextAreaInput()
        chat_interface.button_properties = {
            "check": {"icon": "check", key: callback},
        }
        chat_interface.active_widget.value = "This comes second"
        check_button = chat_interface._input_layout[-1]
        assert check_button.icon == "check"
        check_button.param.trigger("clicks")
        assert chat_interface.objects[0].object == "Checking if this works"

    def test_button_properties_new_callback_and_post_callback(self, chat_interface):
        def pre_callback(instance, event):
            instance.send("1", respond=False)

        def post_callback(instance, event):
            instance.send("2", respond=False)

        chat_interface.widgets = TextAreaInput()
        chat_interface.button_properties = {
            "check": {"callback": pre_callback, "post_callback": post_callback},
        }
        check_button = chat_interface._input_layout[-1]
        check_button.param.trigger("clicks")
        assert chat_interface.objects[0].object == "1"
        assert chat_interface.objects[1].object == "2"

    def test_button_properties_default_callback_and_post_callback(self, chat_interface):
        def post_callback(instance, event):
            instance.send("This should show", respond=False)

        chat_interface.button_properties = {
            "clear": {"post_callback": post_callback},
        }
        clear_button = chat_interface._input_layout[-1]
        chat_interface.send("This shouldn't show up!", respond=False)
        clear_button.param.trigger("clicks")
        assert chat_interface.objects[0].object == "This should show"

    def test_button_properties_send_with_callback_no_duplicate(self, chat_interface):
        def post_callback(instance, event):
            instance.send("This should show", respond=False)

        chat_interface.widgets = TextAreaInput()
        chat_interface.button_properties = {
            "send": {"post_callback": post_callback},
        }
        chat_interface.active_widget.value = "This is it!"
        send_button = chat_interface._input_layout[1]
        send_button.param.trigger("clicks")
        assert chat_interface.objects[0].object == "This is it!"
        assert chat_interface.objects[1].object == "This should show"
        assert len(chat_interface.objects) == 2

    def test_button_properties_new_button_missing_callback(self, chat_interface):
        chat_interface.widgets = TextAreaInput()
        with pytest.raises(ValueError, match="A 'callback' key is required for"):
            chat_interface.button_properties = {
                "check": {"icon": "check"},
            }

    def test_button_properties_update_default(self, chat_interface):
        def callback(instance, event):
            instance.send("This comes first", respond=False)

        chat_interface.widgets = TextAreaInput()
        chat_interface.button_properties = {
            "send": {"icon": "check", "callback": callback},
        }
        chat_interface.active_widget.value = "This comes second"
        send_button = chat_interface._input_layout[1]
        assert send_button.icon == "check"
        send_button.param.trigger("clicks")
        assert chat_interface.objects[0].object == "This comes first"
        assert chat_interface.objects[1].object == "This comes second"

    def test_button_properties_update_default_icon(self, chat_interface):
        chat_interface.widgets = TextAreaInput()
        chat_interface.button_properties = {
            "send": {"icon": "check"},
        }
        chat_interface.active_widget.value = "Test test"
        send_button = chat_interface._input_layout[1]
        assert send_button.icon == "check"
        send_button.param.trigger("clicks")
        assert chat_interface.objects[0].object == "Test test"

    def test_button_properties_update_callback_and_post_callback(self, chat_interface):
        def pre_callback(instance, event):
            instance.send("1", respond=False)

        def post_callback(instance, event):
            instance.send("3", respond=False)

        chat_interface.widgets = TextAreaInput()
        chat_interface.active_widget.value = "2"
        chat_interface.button_properties = {
            "send": {"callback": pre_callback, "post_callback": post_callback},
        }
        send_button = chat_interface._input_layout[1]
        send_button.param.trigger("clicks")
        assert chat_interface.objects[0].object == "1"
        assert chat_interface.objects[1].object == "2"
        assert chat_interface.objects[2].object == "3"

    def test_custom_js_no_code(self):
        chat_interface = ChatInterface()
        with pytest.raises(ValueError, match="A 'code' key is required for"):
            chat_interface.button_properties={
                "help": {
                    "icon": "help",
                    "js_on_click": {
                        "args": {"chat_input": chat_interface.active_widget},
                    },
                },
            }

    def test_manual_user(self):
        chat_interface = ChatInterface(user="New User")
        assert chat_interface.user == "New User"
        chat_interface.send("Test")
        assert chat_interface.objects[0].user == "New User"

    def test_stream_chat_message(self, chat_interface):
        chat_interface.stream(ChatMessage("testeroo", user="useroo", avatar="avataroo"))
        chat_message = chat_interface.objects[0]
        assert chat_message.user == "useroo"
        assert chat_message.avatar == "avataroo"
        assert chat_message.object == "testeroo"

    def test_stream_chat_message_error_passed_user(self, chat_interface):
        with pytest.raises(ValueError, match="Cannot set user or avatar"):
            chat_interface.stream(ChatMessage(
                "testeroo", user="useroo", avatar="avataroo",
            ), user="newuser")

    def test_stream_chat_message_error_passed_avatar(self, chat_interface):
        with pytest.raises(ValueError, match="Cannot set user or avatar"):
            chat_interface.stream(ChatMessage(
                "testeroo", user="useroo", avatar="avataroo",
            ), avatar="newavatar")

    async def test_nested_disabled(self, chat_interface):
        PERSON_1 = "Happy User"
        PERSON_2 = "Excited User"
        PERSON_3 = "Passionate User"

        async def callback(contents: str, user: str, instance: ChatInterface):
            await asyncio.sleep(0.1)
            if user == "User":
                instance.send(
                    f"Hey, {PERSON_2}! Did you hear the user?",
                    user=PERSON_1,
                    avatar="😊",
                    respond=True,  # This is the default, but it's here for clarity
                )
            elif user == PERSON_1:
                user_message = instance.objects[-2]
                user_contents = user_message.object
                yield ChatMessage(
                    f'Yeah, they said "{user_contents}"! Did you also hear {PERSON_3}?',
                    user=PERSON_2,
                    avatar="😄",
                )
                instance.respond()
            elif user == PERSON_2:
                instance.send(
                    'Yup, I heard!',
                    user=PERSON_3,
                    avatar="😆",
                    respond=False,
                )

        chat_interface.callback = callback
        chat_interface.send("Hello")
        await async_wait_until(lambda: chat_interface.objects[-1].object == "Hey, Excited User! Did you hear the user?")
        assert chat_interface.disabled
        await async_wait_until(lambda: chat_interface.objects[-1].object == "Yup, I heard!")
        await asyncio.sleep(0.2)  # give a little time for enabling
        assert not chat_interface.disabled

    def test_prevent_stream_override_message_user_avatar(self, chat_interface):
        msg = chat_interface.send("Hello", user="Welcoming User", avatar="👋")
        chat_interface.stream("New Hello", message=msg)
        assert msg.user == "Welcoming User"
        assert msg.avatar == "👋"

class TestChatInterfaceWidgetsSizingMode:
    def test_none(self):
        chat_interface = ChatInterface()
        assert chat_interface.sizing_mode == "stretch_width"
        assert chat_interface._chat_log.sizing_mode == "stretch_width"
        assert chat_interface._input_layout.sizing_mode == "stretch_width"
        assert chat_interface._input_layout[0].sizing_mode == "stretch_width"

    def test_fixed(self):
        chat_interface = ChatInterface(sizing_mode="fixed")
        assert chat_interface.sizing_mode == "fixed"
        assert chat_interface._chat_log.sizing_mode == "fixed"
        assert chat_interface._input_layout.sizing_mode == "stretch_width"
        assert chat_interface._input_layout[0].sizing_mode == "stretch_width"

    def test_stretch_both(self):
        chat_interface = ChatInterface(sizing_mode="stretch_both")
        assert chat_interface.sizing_mode == "stretch_both"
        assert chat_interface._chat_log.sizing_mode == "stretch_both"
        assert chat_interface._input_layout.sizing_mode == "stretch_width"
        assert chat_interface._input_layout[0].sizing_mode == "stretch_width"

    def test_stretch_width(self):
        chat_interface = ChatInterface(sizing_mode="stretch_width")
        assert chat_interface.sizing_mode == "stretch_width"
        assert chat_interface._chat_log.sizing_mode == "stretch_width"
        assert chat_interface._input_layout.sizing_mode == "stretch_width"
        assert chat_interface._input_layout[0].sizing_mode == "stretch_width"

    def test_stretch_height(self):
        chat_interface = ChatInterface(sizing_mode="stretch_height")
        assert chat_interface.sizing_mode == "stretch_height"
        assert chat_interface._chat_log.sizing_mode == "stretch_height"
        assert chat_interface._input_layout.sizing_mode == "stretch_width"
        assert chat_interface._input_layout[0].sizing_mode == "stretch_width"

    def test_scale_both(self):
        chat_interface = ChatInterface(sizing_mode="scale_both")
        assert chat_interface.sizing_mode == "scale_both"
        assert chat_interface._chat_log.sizing_mode == "scale_both"
        assert chat_interface._input_layout.sizing_mode == "stretch_width"
        assert chat_interface._input_layout[0].sizing_mode == "stretch_width"

    def test_scale_width(self):
        chat_interface = ChatInterface(sizing_mode="scale_width")
        assert chat_interface.sizing_mode == "scale_width"
        assert chat_interface._chat_log.sizing_mode == "scale_width"
        assert chat_interface._input_layout.sizing_mode == "stretch_width"
        assert chat_interface._input_layout[0].sizing_mode == "stretch_width"

    def test_scale_height(self):
        chat_interface = ChatInterface(sizing_mode="scale_height")
        assert chat_interface.sizing_mode == "scale_height"
        assert chat_interface._chat_log.sizing_mode == "scale_height"
        assert chat_interface._input_layout.sizing_mode == "stretch_width"
        assert chat_interface._input_layout[0].sizing_mode == "stretch_width"



"""
Before running this script make sure you have

```bash
git pull
pip install -e . -U
python setup.py develop
```

Then run the script to create/ update the notebooks for Panelite

```bash
python scripts/generate_panelite_content.py
```

Follow the steps in the Panel Jupyterlite guide including

```bash
cd lite
pip install -r requirements.txt -U
jupyter lite build --output-dir ./dist
```

Start the http server

```bash
cd dist
python -m http.server
```

Reset your files as described in https://github.com/jupyterlite/jupyterlite/issues/9#issuecomment-875870620

Run the tests

```bash
pytest scripts/test_panelite/test_panelite.py --headed
```
"""
import pathlib

import pytest

from playwright.sync_api import Page, expect

from scripts.test_panelite.notebooks_with_panelite_issues import (
    NOTEBOOK_ISSUES,
)
from scripts.test_panelite.utils import Retrier

URL = "https://panelite.holoviz.org"
URL = "http://localhost:8000"

PANEL_BASE = pathlib.Path(__file__).parent.parent.parent
FILES = PANEL_BASE / 'lite' / 'files'

def get_panelite_nb_paths():
    nbs = list(FILES.glob('*/*/*.ipynb')) + list(FILES.glob('*/*.*'))
    for nb in nbs:
        path = str(nb).replace("\\", "/").split("files/")[-1]
        if path.endswith(".ipynb") and ".ipynb_checkpoints" not in path:
            yield path
PATHS_WITH_NOTHING_TO_TEST = [
    "gallery/demos/attractors.ipynb",
    "gallery/demos/gapminders.ipynb",
    "gallery/demos/glaciers.ipynb",
    "gallery/demos/nyc_taxi.ipynb",
    "gallery/demos/portfolio-optimizer.ipynb",
]
PATHS = list(path for path in get_panelite_nb_paths() if path not in PATHS_WITH_NOTHING_TO_TEST)
PATHS_WITHOUT_ISSUES = list(path for path in PATHS if path not in NOTEBOOK_ISSUES)
PATHS_WITH_ISSUES = list(path for path in PATHS if path in NOTEBOOK_ISSUES)

def _wait_for_notebook_tab(page, notebook):
    tabs = page.get_by_role("tablist")
    tabs.get_by_text(notebook)

def _select_kernel(page):
    page.get_by_role("button", name="Python (Pyodide)").click()
    page.get_by_role("button", name="Select").filter(has_text="Select").click()

def _run_all_cells(page):
    page.get_by_role("menuitem", name="Run").get_by_text("Run").click()
    page.get_by_text("Run All Cells").nth(2).click()

class RunCells(Exception):
    """Raised if an exception is raised while running all cells"""

def run_notebook(path, page):
    """Opens the notebook and runs all cells"""
    url =URL + f"/lab/index.html?path={path}"
    notebook = path.split("/")[-1]

    retrier = Retrier(retries=3, delay=1)
    while not retrier.accomplished:
        with retrier:
            page.goto(url)
            _wait_for_notebook_tab(page, notebook)
            _select_kernel(page)
            _run_all_cells(page)

            # To wait for * makes tests fail faster
            # The data-status attribute cannot be expected to change within the default timeout
            busy_cells = page.get_by_text("[*]:")
            expect(busy_cells).not_to_have_count(0)

            execution_indicator = page.locator('.jp-Notebook-ExecutionIndicator')
            expect(execution_indicator).to_have_attribute("data-status", "unknown", timeout=10000)
            expect(execution_indicator).to_have_attribute("data-status", "busy", timeout=10000)
            expect(execution_indicator).to_have_attribute("data-status", "idle", timeout=45000)


@pytest.mark.parametrize(["path"], [(path,) for path in PATHS_WITHOUT_ISSUES])
def test_notebooks_without_issues(path, page: Page):
    run_notebook(path, page)
    cells_with_err = page.locator('div[data-mime-type="application/vnd.jupyter.stderr"]')
    if cells_with_err.count()>0:
        raise RunCells(cells_with_err.first.inner_text())

@pytest.mark.xfail
@pytest.mark.parametrize(["path"], [(path,) for path in PATHS_WITH_ISSUES])
def test_notebooks_with_issues(path, page: Page):
    run_notebook(path, page)
    cells_with_err = page.locator('div[data-mime-type="application/vnd.jupyter.stderr"]')
    if cells_with_err.count()>0:
        raise RunCells(cells_with_err.first.inner_text())



from bokeh.models import ImportedStyleSheet

from panel.io.resources import CDN_DIST
from panel.theme.bootstrap import Bootstrap, Inherit
from panel.viewable import Viewable
from panel.widgets import TextInput


class CustomBootstrap(Bootstrap):

    modifiers = {
        Viewable: {
            'stylesheets': [Inherit, 'test.css']
        }
    }

def test_subclassed_inheritance():
    widget = TextInput()

    params, _ = CustomBootstrap().params(widget)
    assert 'stylesheets' in params
    assert len(params['stylesheets']) == 5
    s1, s2, s3, s4, s5 = params['stylesheets']
    assert isinstance(s1, ImportedStyleSheet)
    assert s1.url == f'{CDN_DIST}bundled/bootstrap5/css/bootstrap.min.css'
    assert isinstance(s2, ImportedStyleSheet)
    assert s2.url == f'{CDN_DIST}bundled/theme/default.css'
    assert isinstance(s3, ImportedStyleSheet)
    assert s3.url == f'{CDN_DIST}bundled/theme/bootstrap_default.css'
    assert isinstance(s4, ImportedStyleSheet)
    assert s4.url == f'{CDN_DIST}bundled/theme/bootstrap.css'
    assert s5 == '.bk-input {\n  color: red;\n}\n'


def test_subclassed_inheritance_server(server_document):
    widget = TextInput()

    params, _ = CustomBootstrap().params(widget)
    assert 'stylesheets' in params
    assert len(params['stylesheets']) == 5
    s1, s2, s3, s4, s5 = params['stylesheets']
    assert isinstance(s1, ImportedStyleSheet)
    assert s1.url == f'{CDN_DIST}bundled/bootstrap5/css/bootstrap.min.css'
    assert isinstance(s2, ImportedStyleSheet)
    assert s2.url == f'{CDN_DIST}bundled/theme/default.css'
    assert isinstance(s3, ImportedStyleSheet)
    assert s3.url == f'{CDN_DIST}bundled/theme/bootstrap_default.css'
    assert isinstance(s4, ImportedStyleSheet)
    assert s4.url == f'{CDN_DIST}bundled/theme/bootstrap.css'
    assert isinstance(s5, ImportedStyleSheet)
    assert s5.url == 'components/test_subclass/CustomBootstrap/modifiers/test.css'



import time


class MaxRetriesExceeded(Exception):
     """Raised if the maximum number of retries has been exceeded"""

class Retrier:

    def __init__(self, retries=3, delay=0):
         self.max_retries = retries
         self.delay=delay

         self.retries = 0
         self.accomplished = False

    def __enter__(self):
         return self

    def __exit__(self, exc, value, traceback):
         if not exc:
             self.accomplished = True
             return True
         self.retries += 1
         time.sleep(self.delay)
         if self.retries >= self.max_retries:
             raise MaxRetriesExceeded from exc
         return True



import pathlib
import subprocess
import sys

from panel.custom import JSComponent

CWD = pathlib.Path(__file__).parent


class JSTestComponent(JSComponent):

    _esm = "export function render() { console.log('foo') }"


def test_compile_component(py_file):
    cmd = [sys.executable, "-m", "panel", "compile", "panel.tests.command.test_compile:JSTestComponent", "--unminified"]
    p = subprocess.Popen(cmd, shell=False, cwd=CWD)
    p.wait()

    bundle = CWD / "JSTestComponent.bundle.js"
    try:
        assert bundle.is_file()
        assert 'function render() {\n  console.log("foo");\n}' in bundle.read_text()
    finally:
        bundle.unlink()



from __future__ import annotations

import asyncio
import hashlib
import importlib
import inspect
import os
import pathlib
import sys
import textwrap

from collections import defaultdict
from functools import partial
from typing import (
    TYPE_CHECKING, Any, Callable, ClassVar, Literal, Mapping, Optional,
)

import param

from param.parameterized import ParameterizedMetaclass

from .config import config
from .io.datamodel import construct_data_model
from .io.resources import component_resource_path
from .io.state import state
from .models import (
    AnyWidgetComponent as _BkAnyWidgetComponent,
    ReactComponent as _BkReactComponent, ReactiveESM as _BkReactiveESM,
)
from .models.esm import DataEvent, ESMEvent
from .models.reactive_html import DOMEvent
from .pane.base import PaneBase  # noqa
from .reactive import (  # noqa
    Reactive, ReactiveCustomBase, ReactiveHTML, ReactiveMetaBase,
)
from .util import camel_to_kebab, classproperty
from .util.checks import import_available
from .viewable import (  # noqa
    Child, Children, Layoutable, Viewable, Viewer, is_viewable_param,
)
from .widgets.base import WidgetBase  # noqa

if TYPE_CHECKING:
    from bokeh.document import Document
    from bokeh.events import Event
    from bokeh.model import Model
    from pyviz_comms import Comm

    ExportSpec = dict[str, list[str | tuple[str, ...]]]


class PyComponent(Viewable, Layoutable):
    '''
    The `PyComponent` combines the convenience of `Viewer` components
    that allow creating custom components by declaring a `__panel__`
    method with the ability of controlling layout and styling
    related parameters directly on the class. Internally the
    `PyComponent` will forward layout parameters to the underlying
    object, which is created lazily on render.

    Reference: https://panel.holoviz.org/reference/custom_components/PyComponent.html

    ```python
    import panel as pn
    import param

    pn.extension()

    class CounterButton(pn.custom.PyComponent, pn.widgets.WidgetBase):

        value = param.Integer(default=0)

        def __panel__(self):
            return pn.widgets.Button(
                name=self._button_name, on_click=self._on_click
            )

        def _on_click(self, event):
            self.value += 1

        @param.depends("value")
        def _button_name(self):
            return f"count is {self.value}"

    CounterButton().servable()
    ```
    '''

    def __init__(self, **params):
        super().__init__(**params)
        self._view__ = None
        self._changing__ = {}
        self.param.watch(self._sync__view, [p for p in Layoutable.param if p != 'name'])

    def _sync__view(self, *events):
        if not events or self._view__ is None:
            return
        target = self._view__ if events[0].obj is self else self
        params = {
            e.name: e.new for e in events if e.name not in self._changing__
            or self._changing__[e.name] is not e.new
        }
        if not params:
            return
        try:
            self._changing__.update(params)
            with param.parameterized._syncing(target, list(params)):
                target.param.update(params)
        finally:
            for p in params:
                if p in self._changing__:
                    del self._changing__[p]

    def _cleanup(self, root: Model | None = None) -> None:
        if self._view__ is None:
            return
        super()._cleanup(root)
        if root and root.ref['id'] in self._models:
            del self._models[root.ref['id']]
        self._view__._cleanup(root)

    def _create__view(self):
        from .pane import panel
        from .param import ParamMethod

        if hasattr(self.__panel__, "_dinfo"):
            view = ParamMethod(self.__panel__, lazy=True)
        else:
            view = panel(self.__panel__())
        self._view__ = view
        params = view.param.values()
        overrides, sync = {}, {}
        for p in Layoutable.param:
            if p != 'name' and view.param[p].default != params[p]:
                overrides[p] = params[p]
            elif p != 'name':
                sync[p] = getattr(self, p)
        view.param.watch(self._sync__view, [p for p in Layoutable.param if p != 'name'])
        self.param.update(overrides)
        with param.parameterized._syncing(view, list(sync)):
            view.param.update(sync)

    def _get_model(
        self, doc: Document, root: Optional['Model'] = None,
        parent: Optional['Model'] = None, comm: Optional[Comm] = None
    ) -> 'Model':
        if self._view__ is None:
            self._create__view()
        model = self._view__._get_model(doc, root, parent, comm)
        root = model if root is None else root
        self._models[root.ref['id']] = (model, parent)
        return model

    def select(
        self, selector: Optional[type | Callable[Viewable, bool]] = None
    ) -> list[Viewable]:
        return super().select(selector) + self._view__.select(selector)



class ReactiveESMMetaclass(ReactiveMetaBase):

    def __init__(mcs, name: str, bases: tuple[type, ...], dict_: Mapping[str, Any]):
        mcs.__original_doc__ = mcs.__doc__
        ParameterizedMetaclass.__init__(mcs, name, bases, dict_)

        # Create model with unique name
        ReactiveMetaBase._name_counter[name] += 1
        model_name = f'{name}{ReactiveMetaBase._name_counter[name]}'
        ignored = [p for p in Reactive.param if not issubclass(type(mcs.param[p].owner), ReactiveESMMetaclass)]
        mcs._data_model = construct_data_model(
            mcs, name=model_name, ignore=ignored, extras={'esm_constants': param.Dict}
        )


class ReactiveESM(ReactiveCustomBase, metaclass=ReactiveESMMetaclass):
    '''
    The `ReactiveESM` classes allow you to create custom Panel
    components using HTML, CSS and/ or Javascript and without the
    complexities of Javascript build tools.

    A `ReactiveESM` subclass provides bi-directional syncing of its
    parameters with arbitrary HTML elements, attributes and
    properties. The key part of the subclass is the `_esm`
    variable. Use this to define a `render` function as shown in the
    example below.

    ```python
    import panel as pn
    import param

    pn.extension()

    class CounterButton(pn.custom.ReactiveESM):

        value = param.Integer()

        _esm = """
        export function render({ model }) {
            let btn = document.createElement("button");
            btn.innerHTML = `count is ${model.value}`;
            btn.addEventListener("click", () => {
                model.value += 1
            });
            model.on('value', () => {
                btn.innerHTML = `count is ${model.value}`;
            })
            return btn
        }
        """

    CounterButton().servable()
    ```
    '''

    _bokeh_model = _BkReactiveESM

    _bundle: ClassVar[str | os.PathLike | None] = None

    _constants: ClassVar[dict[str, Any]] = {}

    _esm: ClassVar[str | os.PathLike] = ""

    # Specifies exports to make available to JS in a bundled file
    # 1. Default export: "<export>"
    # 2. Import all (`* as`): "*<export>"
    # 3. Named export (`{ <export>, ... }`): ("<export>", ...)
    _exports__: ClassVar[ExportSpec] = {}

    _importmap: ClassVar[dict[Literal['imports', 'scopes'], dict[str,str]]] = {}

    __abstract = True

    def __init__(self, **params):
        super().__init__(**params)
        self._watching_esm = False
        self._event__callbacks = defaultdict(list)
        self._msg__callbacks = []

    @classproperty
    def _module_path(cls):
        if hasattr(cls, '__path__'):
            return pathlib.Path(cls.__path__)
        try:
            return pathlib.Path(inspect.getfile(cls)).parent
        except (OSError, TypeError, ValueError):
            if not isinstance(cls._bundle, pathlib.PurePath):
                return

    @classproperty
    def _bundle_path(cls) -> os.PathLike | None:
        if config.autoreload and cls._esm:
            return
        mod_path = cls._module_path
        if mod_path is None:
            return
        if cls._bundle:
            for scls in cls.__mro__:
                if issubclass(scls, ReactiveESM) and cls._bundle == scls._bundle:
                    cls = scls
            mod_path = cls._module_path
            bundle = cls._bundle
            if isinstance(bundle, pathlib.PurePath):
                return bundle
            elif bundle.endswith('.js'):
                bundle_path = mod_path / bundle
                if bundle_path.is_file():
                    return bundle_path
            raise ValueError(
                f'Could not resolve {cls.__name__}._bundle: {cls._bundle}. Ensure '
                'you provide either a string with a relative or absolute '
                'path or a Path object to a .js file extension.'
            )

        # Attempt resolving bundle for this component specifically
        path = mod_path / f'{cls.__name__}.bundle.js'
        if path.is_file():
            return path

        # Attempt to resolve bundle in current module and parent modules
        module = cls.__module__
        modules = module.split('.')
        for i in reversed(range(len(modules))):
            submodule = '.'.join(modules[:i+1])
            try:
                mod = importlib.import_module(submodule)
            except (ModuleNotFoundError, ImportError):
                continue
            if not hasattr(mod, '__file__'):
                continue
            submodule_path = pathlib.Path(mod.__file__).parent
            path = submodule_path / f'{submodule}.bundle.js'
            if path.is_file():
                return path

        if module in sys.modules:
            module = os.path.basename(sys.modules[module].__file__).replace('.py', '')
            path = mod_path / f'{module}.bundle.js'
            return path if path.is_file() else None
        return None

    @classmethod
    def _esm_path(cls, compiled: bool = True) -> os.PathLike | None:
        if compiled or not cls._esm:
            bundle_path = cls._bundle_path
            if bundle_path:
                return bundle_path
        esm = cls._esm
        if isinstance(esm, pathlib.PurePath):
            return esm
        elif not esm.endswith(('.js', '.jsx', '.ts', '.tsx')):
            return
        try:
            if hasattr(cls, '__path__'):
                mod_path = cls.__path__
            else:
                mod_path = pathlib.Path(inspect.getfile(cls)).parent
            esm_path = mod_path / esm
            if esm_path.is_file():
                return esm_path
        except (OSError, TypeError, ValueError):
            pass
        return

    @classmethod
    def _render_esm(cls, compiled: bool | Literal['compiling'] = True, server: bool = False):
        esm_path = cls._esm_path(compiled=compiled is True)
        if esm_path:
            if esm_path == cls._bundle_path and cls.__module__ in sys.modules and server:
                base_cls = cls
                for scls in cls.__mro__[1:]:
                    if not issubclass(scls, ReactiveESM):
                        continue
                    if esm_path == scls._esm_path(compiled=compiled is True):
                        base_cls = scls
                esm = component_resource_path(base_cls, '_bundle_path', esm_path)
                if config.autoreload:
                    modified = hashlib.sha256(str(esm_path.stat().st_mtime).encode('utf-8')).hexdigest()
                    esm += f'?{modified}'
            else:
                esm = esm_path.read_text(encoding='utf-8')
        else:
            esm = cls._esm
        esm = textwrap.dedent(esm)
        return esm

    def _cleanup(self, root: Model | None) -> None:
        if root:
            ref = root.ref['id']
            if ref in self._models:
                model, _ = self._models[ref]
                for child in model.children:
                    children = getattr(self, child)
                    if isinstance(children, Viewable):
                        children = [children]
                    if isinstance(children, list):
                        for child in children:
                            if isinstance(child, Viewable):
                                child._cleanup(root)
        super()._cleanup(root)
        if not self._models and self._watching_esm:
            self._watching_esm.set()
            if self._watching_esm in state._watch_events:
                state._watch_events.remove(self._watching_esm)
            self._watching_esm = False

    async def _watch_esm(self):
        import watchfiles
        path = self._esm_path(compiled=False)
        async for _ in watchfiles.awatch(path, stop_event=self._watching_esm):
            self._update_esm()

    def _update_esm(self):
        for ref, (model, _) in self._models.copy().items():
            if ref not in state._views:
                continue
            doc = state._views[ref][2]
            is_session = doc.session_context and doc.session_context.server_context
            esm = self._render_esm(not config.autoreload, server=is_session)
            if esm == model.esm:
                continue
            self._apply_update({}, {'esm': esm}, model, ref)

    @property
    def _linked_properties(self) -> list[str]:
        return [p for p in self._data_model.properties() if p not in ('js_property_callbacks',)]

    def _get_properties(self, doc: Document) -> dict[str, Any]:
        props = super()._get_properties(doc)
        cls = type(self)
        data_params = {}
        # Split data model properties from ESM model properties
        # Note that inherited parameters are generally treated
        # as ESM model properties unless their type has changed
        ignored = [
            p for p in Reactive.param
            if not issubclass(cls.param[p].owner, ReactiveESM) or
            (p in Viewable.param and p != 'name' and type(Reactive.param[p]) is type(cls.param[p]))
        ]
        for k, v in self.param.values().items():
            p = self.param[k]
            is_viewable = is_viewable_param(p)
            if (k in ignored and k != 'name') or ((p.precedence or 0) < 0) or is_viewable:
                if is_viewable and k in props:
                    props.pop(k)
                continue
            if k in props:
                props.pop(k)
            data_params[k] = v
        bundle_path = self._bundle_path
        importmap = self._process_importmap()
        is_session = False
        if bundle_path:
            is_session = (doc.session_context and doc.session_context.server_context)
            if bundle_path == self._esm_path(not config.autoreload) and cls.__module__ in sys.modules and is_session:
                bundle_hash = 'url'
            else:
                bundle_hash = hashlib.sha256(str(bundle_path).encode('utf-8')).hexdigest()
        else:
            bundle_hash = None
        data_props = self._process_param_change(data_params)
        data_props['esm_constants'] = self._constants
        props.update({
            'bundle': bundle_hash,
            'class_name': camel_to_kebab(cls.__name__),
            'data': self._data_model(**{p: v for p, v in data_props.items() if p not in ignored}),
            'dev': config.autoreload or getattr(self, '_debug', False),
            'esm': self._render_esm(not config.autoreload, server=is_session),
            'importmap': importmap,
            'name': cls.__name__
        })
        return props

    @classmethod
    def _process_importmap(cls):
        return cls._importmap

    def _get_child_model(self, child, doc, root, parent, comm):
        if child is None:
            return None
        ref = root.ref['id']
        if isinstance(child, list):
            return [
                sv._models[ref][0] if ref in sv._models else sv._get_model(doc, root, parent, comm)
                for sv in child
            ]
        elif ref in child._models:
            return child._models[ref][0]
        return child._get_model(doc, root, parent, comm)

    def _get_children(self, data_model, doc, root, parent, comm):
        children = {}
        for k, v in self.param.values().items():
            p = self.param[k]
            if not is_viewable_param(p):
                continue
            children[k] = self._get_child_model(v, doc, root, parent, comm)
        return children

    def _setup_autoreload(self):
        from .config import config
        if not ((config.autoreload or getattr(self, '_debug', False)) and import_available('watchfiles')):
            return
        super()._setup_autoreload()
        if (self._esm_path(compiled=False) and not self._watching_esm):
            self._watching_esm = event = asyncio.Event()
            state._watch_events.append(event)
            state.execute(self._watch_esm)

    def _get_model(
        self, doc: Document, root: Optional[Model] = None,
        parent: Optional[Model] = None, comm: Optional[Comm] = None
    ) -> Model:
        model = self._bokeh_model(**self._get_properties(doc))
        root = root or model
        children = self._get_children(model.data, doc, root, model, comm)
        model.data.update(**children)
        model.children = list(children)
        self._models[root.ref['id']] = (model, parent)
        self._link_props(model.data, self._linked_properties, doc, root, comm)
        self._register_events('dom_event', 'data_event', model=model, doc=doc, comm=comm)
        self._setup_autoreload()
        return model

    def _process_event(self, event: 'Event') -> None:
        if isinstance(event, DataEvent):
            for cb in self._msg__callbacks:
                state.execute(partial(cb, event), schedule=False)
            state.execute(partial(self._handle_msg, event.data), schedule=False)
            return
        elif not isinstance(event, DOMEvent):
            return
        if hasattr(self, f'_handle_{event.node}'):
            getattr(self, f'_handle_{event.node}')(event)
        for cb in self._event__callbacks.get(event.node, []):
            state.execute(partial(cb, event), schedule=False)

    def _update_model(
        self, events: dict[str, param.parameterized.Event], msg: dict[str, Any],
        root: Model, model: Model, doc: Document, comm: Optional[Comm]
    ) -> None:
        model_msg, data_msg  = {}, {}
        for prop, v in list(msg.items()):
            if prop in list(Reactive.param)+['esm', 'importmap']:
                model_msg[prop] = v
            elif prop in model.children:
                continue
            else:
                data_msg[prop] = v
        for name, event in events.items():
            if name not in model.children:
                continue
            new = event.new
            old_objects = event.old if isinstance(event.old, list) else [event.old]
            for old in old_objects:
                if old is None or old is new or (isinstance(new, list) and old in new):
                    continue
                old._cleanup(root)
        if any(e in model.children for e in events):
            children = self._get_children(model.data, doc, root, model, comm)
            data_msg.update(children)
            model_msg['children'] = list(children)
        self._set_on_model(model_msg, root, model)
        self._set_on_model(data_msg, root, model.data)

    def _handle_msg(self, data: any) -> None:
        """
        Message handler for messages sent from the frontend using the
        `model.send_msg` API.

        Arguments
        ---------
        data: any
            Data received from the frontend.
        """

    def _send_msg(self, data: any) -> None:
        """
        Sends data to the frontend which can be observed on the frontend
        with the `model.on_msg("msg:custom", callback)` API.

        Arguments
        ---------
        data: any
            Data to send to the frontend.
        """
        self._send_event(ESMEvent, data=data)

    def on_msg(self, callback: Callable) -> None:
        """
        Registers a callback to be executed when a message event
        containing arbitrary data is received.

        Arguments
        ---------
        event: str
          Name of the DOM event to add an event listener to.
        callback: callable
          A callable which will be given the msg data.
        """
        self._msg__callbacks.append(callback)

    def on_event(self, event: str, callback: Callable) -> None:
        """
        Registers a callback to be executed when the specified DOM
        event is triggered.

        Arguments
        ---------
        event: str
          Name of the DOM event to add an event listener to.
        callback: callable
          A callable which will be given the DOMEvent object.
        """
        self._event__callbacks[event].append(callback)


class JSComponent(ReactiveESM):
    '''
    The `JSComponent` allows you to create custom Panel components
    using Javascript and CSS without the complexities of
    Javascript build tools.

    A `JSComponent` subclass provides bi-directional syncing of its
    parameters with arbitrary HTML elements, attributes and
    properties. The key part of the subclass is the `_esm`
    variable. Use this to define a `render` function as shown in the
    example below.

    Reference: https://panel.holoviz.org/reference/custom_components/JSComponent.html

    ```python
    import panel as pn
    import param

    pn.extension()

    class CounterButton(pn.custom.JSComponent):

        value = param.Integer()

        _esm = """
        export function render({ model }) {
            let btn = document.createElement("button");
            btn.innerHTML = `count is ${model.value}`;
            btn.addEventListener("click", () => {
                model.value += 1
            });
            model.on('value', () => {
                btn.innerHTML = `count is ${model.value}`;
            })
            return btn
        }
        """

    CounterButton().servable()
    ```
    '''

    __abstract = True


class ReactComponent(ReactiveESM):
    '''
    The `ReactComponent` allows you to create custom Panel components
    using React without the complexities of Javascript build tools.

    A `ReactComponent` subclass provides bi-directional syncing of its
    parameters with arbitrary HTML elements, attributes and
    properties. The key part of the subclass is the `_esm`
    variable. Use this to define a `render` function as shown in the
    example below.

    Reference: https://panel.holoviz.org/reference/custom_components/ReactComponent.html

    ```python
    import panel as pn
    import param

    class CounterButton(pn.custom.ReactComponent):

        value = param.Integer()

        _esm = """
        export function render({model}) {
        const [value, setValue] = model.useState("value");
        return (
            <button onClick={e => setValue(value+1)}>
            count is {value}
            </button>
        )
        }
        """

    CounterButton().servable()
    ```
    '''

    __abstract = True

    _bokeh_model = _BkReactComponent

    _react_version = '18.3.1'

    @classproperty
    def _exports__(cls) -> ExportSpec:
        imports = cls._importmap.get('imports', {})
        exports = {
            "react": ["*React"],
            "react-dom/client": [("createRoot",)]
        }
        if any('@mui' in v for v in imports.values()):
            exports.update({
                "@emotion/cache": ["createCache"],
                "@emotion/react": [("CacheProvider",)],
            })
        return exports

    @classmethod
    def _render_esm(cls, compiled: bool | Literal['compiling'] = True, server: bool = False):
        esm = super()._render_esm(compiled=compiled, server=server)
        if compiled == 'compiling':
            esm = 'import * as React from "react"\n' + esm
        return esm

    @classmethod
    def _process_importmap(cls):
        imports = cls._importmap.get('imports', {})
        v_react = cls._react_version
        if config.autoreload:
            pkg_suffix, path_suffix = '?dev', '&dev'
        else:
            pkg_suffix = path_suffix = ''
        imports_with_deps = {
            "react": f"https://esm.sh/react@{v_react}{pkg_suffix}",
            "react/": f"https://esm.sh/react@{v_react}{path_suffix}/",
            "react-dom": f"https://esm.sh/react-dom@{v_react}?deps=react@{v_react}&external=react",
            "react-dom/": f"https://esm.sh/react-dom@{v_react}&deps=react@{v_react}{path_suffix}&external=react/"
        }
        suffix = f'deps=react@{v_react},react-dom@{v_react}&external=react,react-dom'
        if any('@mui' in v for v in imports.values()):
            suffix += ',react-is,@emotion/react'
            imports_with_deps.update({
                "react-is": f"https://esm.sh/react-is@{v_react}&external=react",
                "@emotion/cache": f"https://esm.sh/@emotion/cache?deps=react@{v_react},react-dom@{v_react}",
                "@emotion/react": f"https://esm.sh/@emotion/react?deps=react@{v_react},react-dom@{v_react}&external=react,react-is",
                "@emotion/styled": f"https://esm.sh/@emotion/styled?deps=react@{v_react},react-dom@{v_react}&external=react,react-is",
            })
        for k, v in imports.items():
            if '?' not in v and 'esm.sh' in v:
                if v.endswith('/'):
                    v = f'{v[:-1]}?{suffix}&path=/'
                else:
                    v = f'{v}?{suffix}'
            imports_with_deps[k] = v
        return {
            'imports': imports_with_deps,
            'scopes': cls._importmap.get('scopes', {})
        }

class AnyWidgetComponent(ReactComponent):
    '''
    The `AnyWidgetComponent` allows you to create custom Panel components
    in the style of an AnyWidget component. Specifically this component
    type creates shims that make it possible to reuse AnyWidget ESM code
    as is, without having to adapt the callbacks to use Bokeh APIs.

    Reference: https://panel.holoviz.org/reference/custom_components/AnyWidgetComponent.html

    ```python
    import param
    import panel as pn

    pn.extension()

    class CounterWidget(pn.custom.AnyWidgetComponent):
        _esm = """
        function render({ model, el }) {
        let count = () => model.get("value");
        let btn = document.createElement("button");
        btn.innerHTML = `count is ${count()}`;
        btn.addEventListener("click", () => {
            model.set("value", count() + 1);
            model.save_changes();
        });
        model.on("change:value", () => {
            btn.innerHTML = `count is ${count()}`;
        });
        el.appendChild(btn);
        }
        export default { render };
        """
        value = param.Integer()

    CounterWidget().servable()
    ```
    '''

    __abstract = True

    _bokeh_model = _BkAnyWidgetComponent

    def send(self, msg: dict):
        """
        Sends a custom event containing the provided message to the frontend.

        Arguments
        ---------
        msg: dict
        """
        self._send_msg(msg)



import panel as pn

from panel.custom import JSComponent

pn.extension()

class ConfettiButton(JSComponent):

    _importmap = {
        "imports": {
            "confetti": "https://esm.sh/canvas-confetti@1.6.0",
        }
    }

    _esm = """
    import confetti from "confetti";

    export function render() {
        let btn = document.createElement("button");
        btn.innerHTML = `Click Me`;
        btn.addEventListener("click", () => {
            confetti()
        });
        return btn
    }
    """

ConfettiButton().servable()



"""In this module we test the DeckGL Bokeh Model"""

import json

import pytest

from panel.models.deckgl import DeckGLPlot


@pytest.fixture
def json_input():
    return ('{"initialViewState": {"bearing": -27.36, "latitude": 52.2323, '
            '"longitude": -1.415, "maxZoom": 15, "minZoom": 5, "pitch": 40.5, '
            '"zoom": 6}, "layers": [{"@@type": "HexagonLayer", "autoHighlight": '
            'true, "coverage": 1, "data": "https://raw.githubusercontent.com/'
            'uber-common/deck.gl-data/master/examples/3d-heatmap/heatmap-data.csv", '
            '"elevationRange": [0, 3000], "elevationScale": 50, "extruded": true, '
            '"getPosition": "@@=[lng, lat]", "id": "18a4e022-062c-428f-877f-c8c089472297", '
            '"pickable": true}], "mapStyle": "mapbox://styles/mapbox/dark-v9", '
            '"views": [{"@@type": "MapView", "controller": true}]}')


@pytest.fixture
def mapbox_api_key():
    return (
        "pk.eyJ1IjoibWFyY3Nrb3ZtYWRzZW4iLCJhIjoiY2s1anMzcG5rMDYzazNvcm10NTFybTE4cSJ9."
        "TV1XBgaMfR-iTLvAXM_Iew"
    )


@pytest.fixture
def tooltip():
    return True


def test_constructor(json_input, mapbox_api_key, tooltip):
    # When
    data = json.loads(json_input)
    layers = data.pop('layers')
    view_state = data.pop('initialViewState')
    actual = DeckGLPlot(data=data, layers=layers, initialViewState=view_state,
                        mapbox_api_key=mapbox_api_key, tooltip=tooltip,)
    # Then
    assert actual.data == data
    assert actual.layers == layers
    assert actual.initialViewState == view_state
    assert actual.mapbox_api_key == mapbox_api_key
    assert actual.tooltip == tooltip



"""
Helper script to convert and copy example notebooks into JupyterLite build.
"""
import json
import os
import pathlib
import re
import shutil

from test.notebooks_with_panelite_issues import NOTEBOOK_ISSUES

import nbformat

HERE = pathlib.Path(__file__).parent
PANEL_BASE = HERE.parent.parent
EXAMPLES_DIR = PANEL_BASE / 'examples'
LITE_FILES = PANEL_BASE / 'lite' / 'files'
DOC_DIR = PANEL_BASE / 'doc'
BASE_DEPENDENCIES = []
MINIMUM_VERSIONS = {}

INLINE_DIRECTIVE = re.compile('\{.*\}`.*`\s*')

# Add piplite command to notebooks
with open(DOC_DIR / 'pyodide_dependencies.json', encoding='utf8') as file:
    DEPENDENCIES = json.load(file)

class DependencyNotFound(Exception):
    """Raised if a dependency cannot be found"""

def _notebook_key(nbpath: pathlib.Path):
    nbpath = str(nbpath).replace(os.path.sep, "/")
    return nbpath.replace(str(EXAMPLES_DIR), '').replace(str(DOC_DIR), '')[1:]

def _get_dependencies(nbpath: pathlib.Path):
    key = _notebook_key(nbpath)
    dependencies = DEPENDENCIES.get(key, [])
    if dependencies is None:
        return []
    for name, min_version in MINIMUM_VERSIONS.items():
        if any(name in req for req in dependencies):
            dependencies = [f'{name}>={min_version}' if name in req else req for req in dependencies]
    return BASE_DEPENDENCIES + dependencies

def _to_piplite_install_code(dependencies):
    dependencies = [repr(dep) for dep in dependencies]
    return f"import piplite\nawait piplite.install([{', '.join(dependencies)}])"

def _get_install_code_cell(dependencies):
    source = _to_piplite_install_code(dependencies)
    install = nbformat.v4.new_code_cell(source=source)
    del install['id']
    return install

def _get_issues(key):
    return NOTEBOOK_ISSUES.get(key, [])

def _to_issue_str(issue:str):
    if issue.startswith("https://github.com/") and "/issues/" in issue:
        issue = issue.replace("https://github.com/", "")
        _, library, _, id = issue.split("/")
        if library == "panel":
            return f"#{id}"
        return f"{library}#{id}"
    if issue.startswith("https://gitlab.kitware.com/") and "/issues/" in issue:
        issue = issue.replace("https://gitlab.kitware.com/", "")
        _, library, _, _, id = issue.split("/")
        if library == "panel":
            return f"#{id}"
        return f"{library}#{id}"
    return issue

def _get_info_markdown_cell(nbpath):
    key = _notebook_key(nbpath)
    issues = _get_issues(key)
    if issues:
        issues_str = ", ".join([f'<a href="{issue}">{_to_issue_str(issue)}</a>' for issue in issues])
        source = f"""<div class="alert alert-block alert-danger">
This notebook is not expected to run successfully in <em>Panelite</em>. It has the following known issues: {issues_str}.

Panelite is powered by young technologies like <a href="https://pyodide.org/en/stable/">Pyodide</a> and <a href="https://jupyterlite.readthedocs.io/en/latest/">Jupyterlite</a>. Some browsers may be poorly supported (e.g. mobile or 32-bit versions). If you experience other issues, please <a href="https://github.com/holoviz/panel/issues">report them</a>.
</div>"""
    else:
        source = """<div class="alert alert-block alert-success">
<em>Panelite</em> is powered by young technologies like <a href="https://pyodide.org/en/stable/">Pyodide</a> and <a href="https://jupyterlite.readthedocs.io/en/latest/">Jupyterlite</a>. Some browsers may be poorly supported (e.g. mobile or 32-bit versions). If you experience issues, please <a href="https://github.com/holoviz/panel/issues">report them</a>.
</div>"""
    info = nbformat.v4.new_markdown_cell(source=source)
    del info['id']
    return info

def convert_md_to_nb(
    filehandle, supported_syntax=('{pyodide}',)
) -> str:
    """
    Extracts Panel application code from a Markdown file.
    """
    nb = nbformat.v4.new_notebook()
    cells = nb['cells']
    inblock = False
    block_opener = None
    markdown, code = [], []
    while True:
        line = filehandle.readline()
        if not line:
            # EOF
            break

        # Remove MyST-Directives
        line = INLINE_DIRECTIVE.sub('', line)

        lsline = line.lstrip()
        if inblock:
            if lsline.startswith(block_opener):
                inblock = False
                code[-1] = code[-1].rstrip()
                code_cell = nbformat.v4.new_code_cell(source=''.join(code))
                code.clear()
                cells.append(code_cell)
            else:
                code.append(line)
        elif lsline.startswith("```"):
            num_leading_backticks = len(lsline) - len(lsline.lstrip("`"))
            block_opener = '`'*num_leading_backticks
            syntax = line.strip()[num_leading_backticks:]
            if syntax in supported_syntax:
                if markdown:
                    md_cell = nbformat.v4.new_markdown_cell(source=''.join(markdown))
                    markdown.clear()
                    nb['cells'].append(md_cell)
                inblock = True
            else:
                markdown.append(line)
        else:
            markdown.append(line)
    if markdown:
        md_cell = nbformat.v4.new_markdown_cell(source=''.join(markdown))
        nb['cells'].append(md_cell)
    return nb

def convert_docs():
    mds = (
        list(DOC_DIR.glob('getting_started/*.md')) +
        list(DOC_DIR.glob('explanation/**/*.md')) +
        list(DOC_DIR.glob('how_to/**/*.md'))
    )
    for md in mds:
        out = LITE_FILES / md.relative_to(DOC_DIR).with_suffix('.ipynb')
        out.parent.mkdir(parents=True, exist_ok=True)
        if md.suffix != '.md' or md.name == 'index.md':
            continue
        with open(md, encoding="utf-8") as mdf:
            nb = convert_md_to_nb(mdf)
        dependencies = _get_dependencies(md)
        if dependencies:
            install = _get_install_code_cell(dependencies=dependencies)
            nb['cells'].insert(0, install)
            info = _get_info_markdown_cell(md)
            nb['cells'].insert(0, info)
        with open(out, 'w', encoding='utf-8') as fout:
            nbformat.write(nb, fout)

def copy_examples():
    nbs = list(EXAMPLES_DIR.glob('*/*/*.ipynb')) + list(EXAMPLES_DIR.glob('*/*.*'))
    for nb in nbs:
        if ".ipynb_checkpoints" in str(nb):
            continue
        nbpath = pathlib.Path(nb)
        out = LITE_FILES / nbpath.relative_to(EXAMPLES_DIR)
        out.parent.mkdir(parents=True, exist_ok=True)

        if nb.suffix == '.ipynb':
            dependencies = _get_dependencies(nbpath)

            with open(nb, encoding='utf-8') as fin:
                nb = nbformat.read(fin, 4)
                if dependencies:
                    install = _get_install_code_cell(dependencies=dependencies)
                    nb['cells'].insert(0, install)

                info = _get_info_markdown_cell(nbpath)
                nb['cells'].insert(0, info)
            with open(out, 'w', encoding='utf-8') as fout:
                nbformat.write(nb, fout)
        elif not nb.is_dir():
            shutil.copyfile(nb, out)

def copy_assets():
    shutil.copytree(
        EXAMPLES_DIR / 'assets',
        LITE_FILES / 'assets',
        dirs_exist_ok=True
    )
    shutil.copytree(
        DOC_DIR / '_static',
        LITE_FILES / '_static',
        dirs_exist_ok=True
    )

if __name__=="__main__":
    convert_docs()
    copy_examples()
    copy_assets()



"""
Script that removes large files from bokeh wheel and repackages it
to be included in the NPM bundle.
"""

import argparse
import os
import pathlib
import shutil
import subprocess
import zipfile

try:
    import tomllib
except ModuleNotFoundError:
    # Can be removed after 3.11 is the minimum version
    import tomli as tomllib

from packaging.requirements import Requirement

PANEL_BASE = pathlib.Path(__file__).parent.parent
PACKAGE_INFO = tomllib.loads((PANEL_BASE / "pyproject.toml").read_text())
bokeh_requirement = next(p for p in PACKAGE_INFO['build-system']['requires'] if "bokeh" in p.lower())
bokeh_dev = Requirement(bokeh_requirement).specifier.prereleases

parser = argparse.ArgumentParser()
parser.add_argument("out", default="panel/dist/wheels", nargs="?", help="Output dir")
parser.add_argument(
    "--no-deps",
    action="store_true",
    default=False,
    help="Don't install package dependencies.",
)
parser.add_argument(
    "--verify-clean",
    action="store_true",
    default=False,
    help="Check if panel folder is clean before running.",
)
args = parser.parse_args()

if args.verify_clean:
    # -n dry run, -d directories, -x remove ignored files
    output = subprocess.check_output(["git", "clean", "-nxd", "panel/"])
    if output:
        print(output.decode("utf-8"))
        msg = "Please clean the panel folder before running this script."
        raise RuntimeError(msg)
    else:
        print("panel folder is clean.")

command = ["pip", "wheel", "."]
if bokeh_dev:
    command.append("--pre")

if args.no_deps:
    command.append("--no-deps")
command.extend(["-w", str(PANEL_BASE / "build")])
print("command: ", " ".join(command))

out = PANEL_BASE / args.out
out.mkdir(exist_ok=True)
print("out dir: ", out)

sp = subprocess.Popen(command, env=dict(os.environ, PANEL_LITE="1"))
sp.wait()


panel_wheels = list(PANEL_BASE.glob("build/panel-*-py3-none-any.whl"))
if not panel_wheels:
    raise RuntimeError("Panel wheel not found.")
panel_wheel = sorted(panel_wheels)[-1]

with (
    zipfile.ZipFile(panel_wheel, "r") as zin,
    zipfile.ZipFile(out / os.path.basename(panel_wheel).replace(".dirty", ""), "w") as zout,
):
    for item in zin.infolist():
        filename = item.filename
        if filename.startswith("panel/tests"):
            continue
        buffer = zin.read(filename)
        if bokeh_dev and filename.startswith("panel-") and filename.endswith("METADATA"):
            lines = buffer.decode("utf-8").split("\n")
            lines = [
                f"Requires-Dist: {bokeh_requirement}"
                if line.startswith("Requires-Dist: bokeh")
                else line for line in lines
            ]
            buffer = "\n".join(lines).encode('utf-8')
        zout.writestr(item, buffer)

bokeh_wheels = PANEL_BASE.glob("build/bokeh-*-py3-none-any.whl")

if not bokeh_wheels:
    raise RuntimeError("Bokeh wheel not found.")
bokeh_wheel = sorted(bokeh_wheels)[-1]

zin = zipfile.ZipFile(bokeh_wheel, "r")

zout = zipfile.ZipFile(out / os.path.basename(bokeh_wheel), "w")
exts = [".js", ".d.ts", ".tsbuildinfo"]
for item in zin.infolist():
    filename = item.filename
    buffer = zin.read(filename)
    if not filename.startswith("bokeh/core/_templates") and (
        filename.endswith("bokeh.json") or any(filename.endswith(ext) for ext in exts)
    ):
        continue
    elif filename.startswith("bokeh-") and filename.endswith("METADATA"):
        # Replace tornado dependency with bokeh-sampledata
        buffer = "\n".join(
            [
                line if not (
                    "Requires-Dist:" in line and "tornado" in line
                ) else "Requires-Dist: bokeh-sampledata"
                for line in buffer.decode("utf-8").split("\n")
            ]
        ).encode("utf-8")
    zout.writestr(item, buffer)

zout.close()
zin.close()

print(f"\nWheels where successfully written to {out}")



import pathlib

import pytest

from panel.custom import ReactiveESM
from panel.io.compile import (
    find_module_bundles, packages_from_code, packages_from_importmap,
    replace_imports,
)


class ESM1(ReactiveESM):
    pass


def test_find_module_bundles_as_dotted_module():
    assert find_module_bundles('panel.tests.io.test_compile') == {pathlib.Path(__file__).parent / 'ESM1.bundle.js': [ESM1]}

def test_find_module_bundles_as_path():
    path = pathlib.Path(__file__).parent / 'ESM1.bundle.js'
    bundles = find_module_bundles(__file__)
    assert path in bundles
    assert len(bundles[path]) == 1
    assert bundles[path][0].name == 'ESM1'

def test_packages_from_code_esm_sh():
    code, pkgs = packages_from_code('import * from "https://esm.sh/confetti-canvas@1.0.0"')
    assert code == 'import * from "confetti-canvas"'
    assert pkgs == {'confetti-canvas': '^1.0.0'}

def test_packages_from_code_unpkg():
    code, pkgs = packages_from_code('import * from "https://unpkg.com/esm@3.2.25/esm/loader.js"')
    assert code == 'import * from "esm"'
    assert pkgs == {'esm': '^3.2.25'}

def test_packages_from_code_jsdelivr():
    code, pkgs = packages_from_code('import * as vg from "https://cdn.jsdelivr.net/npm/@uwdata/vgplot@0.8.0/+esm"')
    assert code == 'import * as vg from "@uwdata/vgplot"'
    assert pkgs == {'@uwdata/vgplot': '^0.8.0'}

@pytest.mark.parametrize('dev', ['rc', 'a', 'b'])
def test_packages_from_code_dev(dev):
    code, pkgs = packages_from_code(f'import * from "https://esm.sh/confetti-canvas@1.0.0-{dev}.1"')
    assert code == 'import * from "confetti-canvas"'
    assert pkgs == {'confetti-canvas': f'^1.0.0-{dev}.1'}

@pytest.mark.parametrize('import_code', [
    'import * from "https://esm.sh/chart.js@1.0.0/auto"',
    'import * from "https://esm.sh/chart.js@1.0.0/auto?deps=react"',
    'import * from "https://esm.sh/chart.js@1.0.0&external=react/auto"',
])
def test_packages_from_code_path(import_code):
    code, pkgs = packages_from_code(import_code)
    assert code == 'import * from "chart.js/auto"'
    assert pkgs == {'chart.js': '^1.0.0'}

def test_replace_imports_single_quote():
    assert replace_imports("import {foo} from 'bar'", {'bar': 'pkg-bar'}) == "import {foo} from 'pkg-bar'"

def test_replace_imports_double_quote():
    assert replace_imports('import {foo} from "bar"', {'bar': 'pkg-bar'}) == 'import {foo} from "pkg-bar"'

def test_replace_imports_trailing_slash():
    assert replace_imports('import {foo} from "bar/baz"', {'bar': 'pkg-bar'}) == 'import {foo} from "pkg-bar/baz"'

def test_replace_imports_only_from():
    assert replace_imports('import {bar} from "bar"', {'bar': 'pkg-bar'}) == 'import {bar} from "pkg-bar"'

def test_packages_from_importmap_esm_sh():
    code, packages = packages_from_importmap(
        'import * from "confetti"',
        {'confetti': 'https://esm.sh/confetti-canvas@1.0.0'}
    )
    assert code == 'import * from "confetti-canvas"'
    assert packages == {'confetti-canvas': '^1.0.0'}

def test_packages_from_importmap_unpkg():
    code, pkgs = packages_from_importmap(
        'import * from "esm"',
        {'esm': 'https://unpkg.com/esm@3.2.25/esm/loader.js'}
    )
    assert code == 'import * from "esm"'
    assert pkgs == {'esm': '^3.2.25'}

def test_packages_from_importmap_jsdelivr():
    code, pkgs = packages_from_importmap(
        'import * as vg from "@uwdata/vgplot"',
        {"@uwdata/vgplot": "https://cdn.jsdelivr.net/npm/@uwdata/vgplot@0.8.0/+esm"}
    )
    assert code == 'import * as vg from "@uwdata/vgplot"'
    assert pkgs == {'@uwdata/vgplot': '^0.8.0'}

@pytest.mark.parametrize('dev', ['rc', 'a', 'b'])
def test_packages_from_importmap_dev(dev):
    code, pkgs = packages_from_importmap(
        'import * from "confetti"',
        {'confetti': f'https://esm.sh/confetti-canvas@1.0.0-{dev}.1'}
    )
    assert code == 'import * from "confetti-canvas"'
    assert pkgs == {'confetti-canvas': f'^1.0.0-{dev}.1'}

@pytest.mark.parametrize('url', [
    "https://esm.sh/chart.js@1.0.0/",
    "https://esm.sh/chart.js@1.0.0/?deps=react",
    "https://esm.sh/chart.js@1.0.0&external=react/",
])
def test_packages_from_importmap_path(url):
    code, pkgs = packages_from_importmap('import * from "chart-js/auto"', {'chart-js/': url})
    assert code == 'import * from "chart.js/auto"'
    assert pkgs == {'chart.js': '^1.0.0'}



from __future__ import annotations

import asyncio
import dataclasses
import datetime as dt
import gc
import inspect
import json
import logging
import sys
import threading
import time
import weakref

from contextlib import contextmanager
from functools import partial, wraps
from typing import (
    TYPE_CHECKING, Any, Callable, Iterator,
)

from bokeh.application.application import SessionContext
from bokeh.core.serialization import Serializable
from bokeh.document.document import Document
from bokeh.document.events import (
    ColumnDataChangedEvent, ColumnsPatchedEvent, ColumnsStreamedEvent,
    DocumentChangedEvent, MessageSentEvent, ModelChangedEvent,
)
from bokeh.model.util import visit_immediate_value_references
from bokeh.models import CustomJS

from ..config import config
from ..util import param_watchers
from .loading import LOADING_INDICATOR_CSS_CLASS
from .model import monkeypatch_events  # noqa: F401 API import
from .state import curdoc_locked, state

if TYPE_CHECKING:
    from bokeh.core.enums import HoldPolicyType
    from bokeh.core.has_props import HasProps
    from bokeh.protocol.message import Message
    from bokeh.server.connection import ServerConnection
    from pyviz_comms import Comm

logger = logging.getLogger(__name__)

#---------------------------------------------------------------------
# Private API
#---------------------------------------------------------------------

DISPATCH_EVENTS = (
    ColumnDataChangedEvent, ColumnsPatchedEvent, ColumnsStreamedEvent,
    ModelChangedEvent, MessageSentEvent
)
GC_DEBOUNCE = 5
_WRITE_FUTURES = weakref.WeakKeyDictionary()
_WRITE_MSGS = weakref.WeakKeyDictionary()
_WRITE_BLOCK = weakref.WeakKeyDictionary()

_panel_last_cleanup = None
_write_tasks = []

extra_socket_handlers = {}

@dataclasses.dataclass
class Request:
    headers : dict
    cookies : dict
    arguments : dict


class MockSessionContext(SessionContext):

    def __init__(self, *args, document=None, **kwargs):
        self._document = document
        super().__init__(*args, server_context=None, session_id=None, **kwargs)

    def with_locked_document(self, *args):
        return

    @property
    def session(self):
        return None

    @property
    def destroyed(self) -> bool:
        return False

    @property
    def request(self):
        return Request(headers={}, cookies={}, arguments={})

def _cleanup_task(task):
    if task in _write_tasks:
        _write_tasks.remove(task)

def _dispatch_events(doc: Document, events: list[DocumentChangedEvent]) -> None:
    """
    Handles dispatch of events which could not be processed in
    unlocked decorator.
    """
    for event in events:
        doc.callbacks.trigger_on_change(event)

def _cleanup_doc(doc, destroy=True):
    for callback in doc.session_destroyed_callbacks:
        try:
            callback(None)
        except Exception:
            pass
    if hasattr(doc.callbacks, '_change_callbacks'):
        doc.callbacks._change_callbacks[None] = {}

    # Remove views
    from ..viewable import Viewable
    views = {}
    for ref, (pane, root, vdoc, comm) in list(state._views.items()):
        if vdoc is doc:
            pane._cleanup(root)
            if isinstance(pane, Viewable):
                pane._hooks = []
                for p in pane.select():
                    p._hooks = []
                    param_watchers(p, {})
                    p._documents = {}
                    p._internal_callbacks = {}
            param_watchers(pane, {})
            pane._documents = {}
            pane._internal_callbacks = {}
        else:
            views[ref] = (pane, root, vdoc, comm)
    state._views = views

    # When reusing sessions we must clean up the Panel state but we
    # must **not** destroy the template or the document
    if not destroy:
        return

    # Clean up templates
    if doc in state._templates:
        tmpl = state._templates[doc]
        tmpl._documents = {}
        del state._templates[doc]

    # Destroy document
    doc.destroy(None)

async def _run_write_futures(doc):
    """
    Ensure that all write_message calls are awaited and handled.
    """
    from tornado.websocket import WebSocketClosedError
    futures = _WRITE_FUTURES.pop(doc, [])
    for future in futures:
        try:
            await future
        except WebSocketClosedError:
            logger.warning("Failed sending message as connection was closed")
        except Exception as e:
            logger.warning(f"Failed sending message due to following error: {e}")

def _dispatch_write_task(doc, func, *args, **kwargs):
    """
    Schedules tasks that write messages to the socket.
    """
    try:
        task = asyncio.ensure_future(func(*args, **kwargs))
        _write_tasks.append(task)
        task.add_done_callback(_cleanup_task)
    except RuntimeError:
        doc.add_next_tick_callback(partial(func, *args, **kwargs))

async def _dispatch_msgs(doc):
    """
    Writes messages to a socket, ensuring that the write_lock is not
    set, otherwise re-schedules the write task on the event loop.
    """
    from tornado.websocket import WebSocketHandler
    remaining = {}
    futures = []
    conn_msgs = _WRITE_MSGS.pop(doc, {})
    for conn, msgs in conn_msgs.items():
        socket = conn._socket
        if hasattr(socket, 'write_lock') and socket.write_lock._block._value == 0:
            remaining[conn] = msgs
            continue
        for msg in msgs:
            if isinstance(conn._socket, WebSocketHandler):
                futures += dispatch_tornado(conn, msg=msg)
            elif (socket_type:= type(conn._socket)) in extra_socket_handlers:
                futures += extra_socket_handlers[socket_type](conn, msg=msg)
            else:
                futures += dispatch_django(conn, msg=msg)
    if futures:
        if doc in _WRITE_FUTURES:
            _WRITE_FUTURES[doc] += futures
        else:
            _WRITE_FUTURES[doc] = futures
        await _run_write_futures(doc)
    if not remaining:
        if doc in _WRITE_BLOCK:
            del _WRITE_BLOCK[doc]
        return
    for conn, msgs in remaining.items():
        if doc in _WRITE_MSGS:
            _WRITE_MSGS[doc][conn] = msgs + _WRITE_MSGS[doc].get(conn, [])
        else:
            _WRITE_MSGS[doc] = {conn: msgs}
    await asyncio.sleep(0.01)
    _dispatch_write_task(doc, _dispatch_msgs, doc)

def _garbage_collect():
    if (new_time:= time.monotonic()-_panel_last_cleanup) < GC_DEBOUNCE:
        at = dt.datetime.now() + dt.timedelta(seconds=new_time)
        state.schedule_task('gc.collect', _garbage_collect, at=at)
        return
    gc.collect()

def _destroy_document(self, session):
    """
    Override for Document.destroy() without calling gc.collect directly.
    The gc.collect() call is scheduled as a task, ensuring that when
    multiple documents are destroyed in quick succession we do not
    schedule excessive garbage collection.
    """
    if session is not None:
        self.remove_on_change(session)

    del self._roots
    del self._theme
    del self._template
    self._session_context = None

    self.callbacks.destroy()
    self.models.destroy()

    # Module cleanup without trawling through referrers (as self.modules.destroy() does)
    for module in self.modules._modules:
        # remove the reference from sys.modules
        if module.__name__ in sys.modules:
            del sys.modules[module.__name__]

        # explicitly clear the module contents and the module here itself
        module.__dict__.clear()
        del module
    self.modules._modules = []

    # Clear periodic callbacks
    for cb in state._periodic.get(self, []):
        cb.stop()

    # Clean up pn.state to avoid tasks getting executed on dead session
    for attr in dir(state):
        # _param_watchers is deprecated in Param 2.0 and will raise a warning
        if not attr.startswith('_') or attr == "_param_watchers":
            continue
        state_obj = getattr(state, attr)
        if isinstance(state_obj, weakref.WeakKeyDictionary) and self in state_obj:
            del state_obj[self]

    # Schedule GC
    global _panel_last_cleanup
    _panel_last_cleanup = time.monotonic()
    at = dt.datetime.now() + dt.timedelta(seconds=GC_DEBOUNCE)
    state.schedule_task('gc.collect', _garbage_collect, at=at)

    del self.destroy

#---------------------------------------------------------------------
# Public API
#---------------------------------------------------------------------

def retrigger_events(doc: Document, events: list[DocumentChangedEvent]):
    """
    Applies events that could not be processed previously.
    """
    if doc.callbacks.hold_value:
        doc.callbacks._held_events = events + list(doc.callbacks._held_events)
    else:
        _dispatch_events(doc, events)

def write_events(
    doc: Document,
    connections: list[ServerConnection],
    events: list[DocumentChangedEvent]
):
    from tornado.websocket import WebSocketHandler

    futures = []
    for conn in connections:
        if isinstance(conn._socket, WebSocketHandler):
            futures += dispatch_tornado(conn, events)
        elif (socket_type:= type(conn._socket)) in extra_socket_handlers:
            futures += extra_socket_handlers[socket_type](conn, events)
        else:
            futures += dispatch_django(conn, events)

    if doc in _WRITE_FUTURES:
        _WRITE_FUTURES[doc] += futures
    else:
        _WRITE_FUTURES[doc] = futures

    if state._unblocked(doc):
        _dispatch_write_task(doc, _run_write_futures, doc)
    else:
        doc.add_next_tick_callback(partial(_run_write_futures, doc))

def schedule_write_events(
    doc: Document,
    connections: list[ServerConnection],
    events: list[DocumentChangedEvent]
):
    # Set up write locks
    _WRITE_BLOCK[doc] = True
    _WRITE_MSGS[doc] = msgs = _WRITE_MSGS.get(doc, {})
    # Create messages for remaining events
    for conn in connections:
        # Create a protocol message for any events that cannot be immediately dispatched
        msg = conn.protocol.create('PATCH-DOC', events)
        if conn in msgs:
            msgs[conn].append(msg)
        else:
            msgs[conn] = [msg]
    _dispatch_write_task(doc, _dispatch_msgs, doc)

def create_doc_if_none_exists(doc: Document | None) -> Document:
    curdoc = doc or curdoc_locked()
    if curdoc is None:
        curdoc = Document()
    elif not isinstance(curdoc, Document):
        curdoc = curdoc._doc
    return curdoc

def init_doc(doc: Document | None) -> Document:
    curdoc = create_doc_if_none_exists(doc)
    if not curdoc.session_context:
        return curdoc

    thread_id = threading.get_ident()
    if thread_id:
        state._thread_id_[curdoc] = thread_id

    if config.global_loading_spinner:
        curdoc.js_on_event(
            'document_ready', CustomJS(code=f"""
            const body = document.getElementsByTagName('body')[0]
            body.classList.remove({LOADING_INDICATOR_CSS_CLASS!r}, {config.loading_spinner!r})
            """)
        )

    session_id = curdoc.session_context.id
    sessions = state.session_info['sessions']
    if session_id not in sessions:
        return curdoc

    sessions[session_id].update({
        'started': dt.datetime.now().timestamp()
    })
    curdoc.on_event('document_ready', state._init_session)
    return curdoc

def with_lock(func: Callable) -> Callable:
    """
    Wrap a callback function to execute with a lock allowing the
    function to modify bokeh models directly.

    Arguments
    ---------
    func: callable
      The callable to wrap

    Returns
    -------
    wrapper: callable
      Function wrapped to execute without a Document lock.
    """
    if inspect.iscoroutinefunction(func):
        @wraps(func)
        async def wrapper(*args, **kw):
            return await func(*args, **kw)
    else:
        @wraps(func)
        def wrapper(*args, **kw):
            return func(*args, **kw)
    wrapper.lock = True # type: ignore
    return wrapper

def dispatch_tornado(
    conn: ServerConnection,
    events: list[DocumentChangedEvent] | None = None,
    msg: Message | None = None
):
    from tornado.websocket import WebSocketHandler
    socket = conn._socket
    ws_conn = getattr(socket, 'ws_connection', False)
    if not ws_conn or ws_conn.is_closing(): # type: ignore
        return []
    if msg is None:
        msg = conn.protocol.create('PATCH-DOC', events)
    futures = [
        WebSocketHandler.write_message(socket, msg.header_json),
        WebSocketHandler.write_message(socket, msg.metadata_json),
        WebSocketHandler.write_message(socket, msg.content_json)
    ]
    for buffer in msg._buffers:
        header = json.dumps(buffer.ref)
        payload = buffer.to_bytes()
        futures.extend([
            WebSocketHandler.write_message(socket, header),
            WebSocketHandler.write_message(socket, payload, binary=True)
        ])
    return futures

def dispatch_django(
    conn: ServerConnection,
    events: list[DocumentChangedEvent] | None = None,
    msg: Message | None = None
):
    socket = conn._socket
    if msg is None:
        msg = conn.protocol.create('PATCH-DOC', events)
    futures = [
        socket.send(text_data=msg.header_json),
        socket.send(text_data=msg.metadata_json),
        socket.send(text_data=msg.content_json)
    ]
    for buffer in msg._buffers:
        header = json.dumps(buffer.ref)
        payload = buffer.to_bytes()
        futures.extend([
            socket.send(text_data=header),
            socket.send(binary_data=payload)
        ])
    return futures

@contextmanager
def unlocked(policy: HoldPolicyType = 'combine') -> Iterator:
    """
    Context manager which unlocks a Document and dispatches
    ModelChangedEvents triggered in the context body to all sockets
    on current sessions.

    Arguments
    ---------
    policy: Literal['combine' | 'collect']
        One of 'combine' or 'collect' determining whether events
        setting the same property are combined or accumulated to be
        dispatched when the context manager exits.
    """
    curdoc = state.curdoc
    session_context = getattr(curdoc, 'session_context', None)
    session = getattr(session_context, 'session', None)
    if state._current_thread != state._thread_id and state.loaded and session:
        logger.error(
            "Using the unlocked decorator when running inside a thread "
            "is not safe! Ensure you check that pn.state._current_thread "
            "matches the current thread id."
        )
        yield
        return
    elif (curdoc is None or session is None or not state.loaded or
          state._jupyter_kernel_context):
        yield
        return
    elif curdoc.callbacks.hold_value:
        yield
        monkeypatch_events(curdoc.callbacks._held_events)
        return

    curdoc.hold(policy=policy)
    try:
        yield
    finally:
        # Whether or not there was an error in the body of context manager
        # we may have captured some events. We will dispatch these
        # either by running the write futures, by serializing them
        # as actual messages and scheduling these messages to be written,
        # by having bokeh dispatch them on calling unhold or by
        # scheduling them to be triggered later.
        connections = session._subscribed_connections
        locked = curdoc in _WRITE_MSGS or curdoc in _WRITE_BLOCK
        for conn in connections:
            socket = conn._socket
            if hasattr(socket, 'write_lock') and socket.write_lock._block._value == 0:
                locked = True
                break

        remaining_events, writeable_events = [], []
        events = list(curdoc.callbacks._held_events or [])
        curdoc.callbacks._held_events = []
        monkeypatch_events(events)
        for event in events:
            if isinstance(event, DISPATCH_EVENTS) and not locked:
                writeable_events.append(event)
            else:
                remaining_events.append(event)

        try:
            if writeable_events:
                write_events(curdoc, connections, writeable_events)
        except Exception:
            remaining_events = events
        finally:
            # If for whatever reasons there are still events that couldn't
            # be dispatched we create a protocol message for these immediately
            # and then schedule a task to write the message to the websocket
            # on the next iteration of the event loop. This ensures that
            # the message reflects the event at the time it was generated
            # potentially avoiding issues serializing subsequent models
            # which assume the serializer has previously seen them.
            serializable_events = [e for e in remaining_events if isinstance(e, Serializable)]
            held_events = [e for e in remaining_events if not isinstance(e, Serializable)]
            if serializable_events:
                try:
                    schedule_write_events(curdoc, connections, serializable_events)
                except Exception:
                    # If the serialization fails we let bokeh handle them
                    held_events = remaining_events
            curdoc.callbacks._held_events += held_events

            # Last we attempt to let bokeh handle these remaining events
            # if this also fails we reapply the event at a later point in
            # time. This should not happen but since network writes
            # are fickle we handle this case anyway.
            try:
                retriggered_events = list(curdoc.callbacks._held_events)
                curdoc.unhold()
            except RuntimeError:
                curdoc.add_next_tick_callback(partial(retrigger_events, curdoc, retriggered_events))

@contextmanager
def hold(doc: Document | None = None, policy: HoldPolicyType = 'combine', comm: Comm | None = None):
    """
    Context manager that holds events on a particular Document
    allowing them all to be collected and dispatched when the context
    manager exits. This allows multiple events on the same object to
    be combined if the policy is set to 'combine'.

    Arguments
    ---------
    doc: Document
        The Bokeh Document to hold events on.
    policy: HoldPolicyType
        One of 'combine', 'collect' or None determining whether events
        setting the same property are combined or accumulated to be
        dispatched when the context manager exits.
    comm: Comm
        The Comm to dispatch events on when the context manager exits.
    """
    doc = doc or state.curdoc
    if doc is None:
        yield
        return
    held = doc.callbacks.hold_value
    try:
        if policy is None:
            doc.unhold()
            yield
        else:
            with unlocked(policy=policy):
                if not doc.callbacks.hold_value:
                    doc.hold(policy)
                yield
    finally:
        if held:
            doc.callbacks._hold = held
        else:
            if comm is not None:
                from .notebook import push
                push(doc, comm)
            doc.unhold()

@contextmanager
def immediate_dispatch(doc: Document | None = None):
    """
    Context manager to trigger immediate dispatch of events triggered
    inside the execution context even when Document events are
    currently on hold.

    Arguments
    ---------
    doc: Document
        The document to dispatch events on (if `None` then `state.curdoc` is used).
    """
    doc = doc or state.curdoc

    # Skip if not in a server context
    if not doc or not doc._session_context or not state._unblocked(doc):
        yield
        return

    old_events = doc.callbacks._held_events
    held = doc.callbacks._hold
    doc.callbacks._held_events = []
    doc.callbacks.unhold()
    with unlocked():
        yield
    doc.callbacks._hold = held
    doc.callbacks._held_events = old_events

@contextmanager
def freeze_doc(doc: Document, model: HasProps, properties: dict[str, Any], force: bool = False):
    """
    Freezes the document model references if any of the properties
    are themselves a model.
    """
    if not hasattr(doc, '_roots'):
        dirty_count = 0
    elif force:
        dirty_count = 1
    else:
        dirty_count = 0
        def mark_dirty(_: HasProps):
            nonlocal dirty_count
            dirty_count += 1
        for key, value in properties.items():
            visit_immediate_value_references(getattr(model, key, None), mark_dirty)
            visit_immediate_value_references(value, mark_dirty)
    if dirty_count:
        doc.models._push_freeze()
    yield
    if dirty_count:
        doc.models._pop_freeze()



from __future__ import annotations

import traceback

from typing import ClassVar, Literal, Mapping

import param

from ..io.resources import CDN_DIST
from ..layout import Card, Row
from ..pane.image import ImageBase
from ..pane.markup import HTML, HTMLBasePane, Markdown
from ..pane.placeholder import Placeholder
from ..util import edit_readonly
from ..widgets.indicators import BooleanStatus
from .utils import (
    avatar_lookup, build_avatar_pane, serialize_recursively, stream_to,
)

DEFAULT_STATUS_BADGES = {
    "pending": BooleanStatus(value=False, margin=0, color="primary"),
    "running": BooleanStatus(value=True, margin=0, color="warning"),
    "success": BooleanStatus(value=True, margin=0, color="success"),
    "failed": BooleanStatus(value=True, margin=0, color="danger"),
}


class ChatStep(Card):
    """
    A component that makes it easy to provide status updates and the
    ability to stream updates to both the output(s) and the title.

    Reference: https://panel.holoviz.org/reference/chat/ChatStep.html

    :Example:

    >>> ChatStep("Hello world!", title="Running calculation...', status="running")
    """

    collapsed = param.Boolean(default=False, doc="""
        Whether the contents of the Card are collapsed.""")

    collapsed_on_success = param.Boolean(default=True, doc="""
        Whether to collapse the card on completion.""")

    context_exception = param.ObjectSelector(
        default="raise", objects=["raise", "summary", "verbose", "ignore"], doc="""
        How to handle exceptions raised upon exiting the context manager.
        If "raise", the exception will be raised.
        If "summary", a summary will be sent to the chat step.
        If "verbose", the full traceback will be sent to the chat step.
        If "ignore", the exception will be ignored.
        """)

    success_title = param.String(default=None, doc="""
        Title to display when status is success.""")

    default_badges = param.Dict(default=DEFAULT_STATUS_BADGES, doc="""
        Mapping from status to default status badge; keys must be one of
        'pending', 'running', 'success', 'failed'.
        """)

    default_title = param.String(default="", doc="""
        The default title to display if the other title params are unset.""")

    failed_title = param.String(default=None, doc="""
        Title to display when status is failed.""")

    header = param.Parameter(doc="""
        A Panel component to display in the header bar of the Card.
        Will override the given title if defined.""", readonly=True)

    margin = param.Parameter(default=(5, 5, 5, 10), doc="""
        Allows to create additional space around the component. May
        be specified as a two-tuple of the form (vertical, horizontal)
        or a four-tuple (top, right, bottom, left).""")

    pending_title = param.String(default=None, doc="""
        Title to display when status is pending.""")

    running_title = param.String(default=None, doc="""
        Title to display when status is running.""")

    status = param.Selector(default="pending", objects=[
        "pending", "running", "success", "failed"])

    title = param.String(default="", constant=True, doc="""
        The title of the chat step. Will redirect to default_title on init.
        After, it cannot be set directly; instead use the *_title params.""")

    _rename: ClassVar[Mapping[str, str | None]] = {
        "collapsed_on_success": None,
        "context_exception": None,
        "default_badges": None,
        "default_title": None,
        "pending_title": None,
        "running_title": None,
        "success_title": None,
        "failed_title": None,
        "status": None,
        **Card._rename,
    }

    _stylesheets: ClassVar[list[str]] = [f"{CDN_DIST}css/chat_step.css"]

    def __init__(self, *objects, **params):
        self._instance = None
        self._failed_title = ""  # for cases when contextmanager is invoked twice
        self._avatar_placeholder = Placeholder(css_classes=["step-avatar-container"])

        if params.get("title"):
            if params.get("default_title"):
                raise ValueError("Cannot set both title and default_title.")
            params["default_title"] = params["title"]

        super().__init__(*objects, **params)
        self._render_avatar()
        self._title_pane = HTML(
            self.param.title,
            margin=0,
            css_classes=["step-title"],
        )

        with edit_readonly(self):
            self.header = Row(
                self._avatar_placeholder,
                self._title_pane,
                stylesheets=self._stylesheets + self.param.stylesheets.rx(),
                css_classes=["step-header"],
                margin=(5, 0),
                width=self.width,
                max_width=self.max_width,
                min_width=self.min_width,
                sizing_mode=self.sizing_mode,
            )

    def __enter__(self):
        self.status = "running"
        return self

    def __exit__(self, exc_type, exc_value, tb):
        if exc_type is not None and self.context_exception != "ignore":
            if self.failed_title is None:
                # convert to str to wrap repr in apostrophes
                self._failed_title = f"Error: {exc_type.__name__!r}"
                if self.context_exception == "summary" or self.context_exception == "raise":
                    exc_msg = f"{exc_value}"
                elif self.context_exception == "verbose":
                    exc_msg = "```python\n" + "\n".join(traceback.format_exception(exc_type, exc_value, tb)) + "\n```"
                else:
                    exc_msg = None

                if len(self.objects) == 1:
                    exc_msg = f"\n{exc_msg}"
                self.stream(exc_msg)
            self.status = "failed"
            if self.context_exception == "raise":
                return False
        else:
            self.status = "success"
        return True  # suppress exception if any

    @param.depends("status", "default_badges", watch=True)
    def _render_avatar(self):
        """
        Render the avatar pane as some HTML text or Image pane.
        """
        extra_keys = set(self.default_badges.keys()) - set(DEFAULT_STATUS_BADGES.keys())
        if extra_keys:
            raise ValueError(
                f"Invalid status avatars. Must be one of 'pending', 'running', 'success', 'failed'; "
                f"got {extra_keys}."
            )

        avatar = avatar_lookup(
            self.status,
            None,
            self.default_badges,
            DEFAULT_STATUS_BADGES,
        )
        avatar_pane = build_avatar_pane(avatar, ["step-avatar"])
        self._avatar_placeholder.update(avatar_pane)

    @param.depends(
        "status",
        "default_title",
        "pending_title",
        "running_title",
        "success_title",
        "failed_title",
        watch=True,
        on_init=True,
    )
    def _update_title_on_status(self):
        if self.status == "pending" and self.pending_title is not None:
            title = self.pending_title
        elif self.status == "running" and self.running_title is not None:
            title = self.running_title
        elif self.status == "success" and self.success_title is not None:
            title = self.success_title
        elif self.status == "failed" and self.failed_title is not None:
            title = self.failed_title
        elif self.status == "failed" and self._failed_title:
            title = self._failed_title
        else:
            title = self.default_title

        with param.edit_constant(self):
            self.title = title

    @param.depends("status", "collapsed_on_success", watch=True)
    def _update_collapsed(self):
        if self.status == "success" and self.collapsed_on_success:
            self.collapsed = True

    def stream_title(
        self,
        token: str,
        status: Literal["pending", "running", "success", "failed", "default"] = "running",
        replace: bool = False
    ):
        """
        Stream a token to the title header.

        Arguments:
        ---------
        token : str
            The token to stream.
        status : str
            The status title to stream to, one of 'pending', 'running', 'success', 'failed', or "default".
        replace : bool
            Whether to replace the existing text.
        """
        if replace:
            setattr(self, f"{status}_title", token)
        else:
            original = getattr(self, f"{status}_title") or ""
            setattr(self, f"{status}_title", original + token)

    def stream(self, token: str | None, replace: bool = False):
        """
        Stream a token to the last available string-like object.

        Arguments
        ---------
        token : str
            The token to stream.
        replace : bool
            Whether to replace the existing text.

        Returns
        -------
        Viewable
            The updated message pane.
        """
        if token is None:
            token = ""

        if (
            len(self.objects) == 0 or not isinstance(self.objects[-1], HTMLBasePane) or isinstance(self.objects[-1], ImageBase)
        ):
            message = Markdown(token, css_classes=["step-message"])
            self.append(message)
        else:
            stream_to(self.objects[-1], token, replace=replace)

    def serialize(
        self,
        prefix_with_viewable_label: bool = True,
        prefix_with_container_label: bool = True,
    ) -> str:
        """
        Format the object to a string.

        Arguments
        ---------
        prefix_with_viewable_label : bool
            Whether to include the name of the Viewable, or type
            of the viewable if no name is specified.
        prefix_with_container_label : bool
            Whether to include the name of the container, or type
            of the container if no name is specified.

        Returns
        -------
        str
            The serialized string.
        """
        return serialize_recursively(
            self,
            prefix_with_viewable_label=prefix_with_viewable_label,
            prefix_with_container_label=prefix_with_container_label,
        )

    def __str__(self):
        return self.serialize()



import pytest

pytest.importorskip("playwright")


from panel.layout import Row
from panel.tests.util import serve_component, wait_until
from panel.widgets import (
    ButtonIcon, IntInput, StaticText, ToggleIcon,
)

pytestmark = pytest.mark.ui

SVG = """
<svg xmlns="http://www.w3.org/2000/svg" class="icon icon-tabler icon-tabler-ad-off" width="24" height="24" viewBox="0 0 24 24" stroke-width="2" stroke="currentColor" fill="none" stroke-linecap="round" stroke-linejoin="round"><path stroke="none" d="M0 0h24v24H0z" fill="none"/><path d="M9 5h10a2 2 0 0 1 2 2v10m-2 2h-14a2 2 0 0 1 -2 -2v-10a2 2 0 0 1 2 -2" /><path d="M7 15v-4a2 2 0 0 1 2 -2m2 2v4" /><path d="M7 13h4" /><path d="M17 9v4" /><path d="M16.115 12.131c.33 .149 .595 .412 .747 .74" /><path d="M3 3l18 18" /></svg>
"""  # noqa: E501
ACTIVE_SVG = """
<svg xmlns="http://www.w3.org/2000/svg" class="icon icon-tabler icon-tabler-ad-filled" width="24" height="24" viewBox="0 0 24 24" stroke-width="2" stroke="currentColor" fill="none" stroke-linecap="round" stroke-linejoin="round"><path stroke="none" d="M0 0h24v24H0z" fill="none"/><path d="M19 4h-14a3 3 0 0 0 -3 3v10a3 3 0 0 0 3 3h14a3 3 0 0 0 3 -3v-10a3 3 0 0 0 -3 -3zm-10 4a3 3 0 0 1 2.995 2.824l.005 .176v4a1 1 0 0 1 -1.993 .117l-.007 -.117v-1h-2v1a1 1 0 0 1 -1.993 .117l-.007 -.117v-4a3 3 0 0 1 3 -3zm0 2a1 1 0 0 0 -.993 .883l-.007 .117v1h2v-1a1 1 0 0 0 -1 -1zm8 -2a1 1 0 0 1 .993 .883l.007 .117v6a1 1 0 0 1 -.883 .993l-.117 .007h-1.5a2.5 2.5 0 1 1 .326 -4.979l.174 .029v-2.05a1 1 0 0 1 .883 -.993l.117 -.007zm-1.41 5.008l-.09 -.008a.5 .5 0 0 0 -.09 .992l.09 .008h.5v-.5l-.008 -.09a.5 .5 0 0 0 -.318 -.379l-.084 -.023z" stroke-width="0" fill="currentColor" /></svg>
"""  # noqa: E501


def test_toggle_icon_click(page):
    icon = ToggleIcon()
    serve_component(page, icon)

    # test defaults
    assert icon.icon == "heart"
    assert not icon.value
    icon_element = page.locator(".ti-heart")
    assert icon_element

    events = []

    def cb(event):
        events.append(event)

    icon.param.watch(cb, "value")

    # test icon click updates value
    page.click(".bk-TablerIcon")
    wait_until(lambda: len(events) == 1, page)
    assert icon.value


def test_toggle_icon_width_height(page):
    icon = ToggleIcon(width=100, height=100)
    serve_component(page, icon)

    # test defaults
    assert icon.icon == "heart"
    assert not icon.value
    icon_element = page.locator(".ti-heart")

    wait_until(lambda: icon_element.bounding_box()["width"] == 100)


def test_toggle_icon_size(page):
    icon = ToggleIcon(size="120px")
    serve_component(page, icon)

    # test defaults
    assert icon.icon == "heart"
    assert not icon.value
    icon_element = page.locator(".ti-heart")

    wait_until(lambda: icon_element.bounding_box()["width"] == 120)


def test_toggle_icon_active_icon(page):
    icon = ToggleIcon(icon="thumb-down", active_icon="thumb-up")
    serve_component(page, icon)

    # test defaults
    assert icon.icon == "thumb-down"
    assert not icon.value
    assert page.locator(".thumb-down")

    events = []

    def cb(event):
        events.append(event)

    icon.param.watch(cb, "value")

    # test icon click updates value
    page.click(".bk-TablerIcon")
    wait_until(lambda: len(events) == 1, page)
    assert icon.value
    assert page.locator(".thumb-up")


def test_toggle_icon_update_params_dynamically(page):
    icon = ToggleIcon(icon="thumb-down", active_icon="thumb-up")
    serve_component(page, icon)

    # test defaults
    assert icon.icon == "thumb-down"
    assert not icon.value
    assert page.locator(".thumb-down")

    events = []

    def cb(event):
        events.append(event)

    icon.param.watch(cb, "value")

    # test icon click updates value
    page.click(".bk-TablerIcon")
    wait_until(lambda: len(events) == 1, page)
    assert icon.value
    assert page.locator(".thumb-up")

    # update active icon name
    icon.active_icon = ""
    assert page.locator(".heart-filled")

    # update value and icon name
    icon.value = True
    icon.icon = "heart"
    assert page.locator(".ti-heart")

    # update active icon_name to svg
    icon.active_icon = ACTIVE_SVG
    assert page.locator(".icon-tabler-ad-filled")

    # update size
    icon.size = "8em"
    assert page.locator(".icon-tabler-ad-filled").bounding_box()["width"] >= 96
    icon.size = "1em"
    wait_until(lambda: page.locator(".icon-tabler-ad-filled").bounding_box()["width"] <= 24, page)


def test_toggle_icon_svg(page):
    icon = ToggleIcon(icon=SVG, active_icon=ACTIVE_SVG)
    serve_component(page, icon)

    # test defaults
    assert icon.icon == SVG
    assert not icon.value
    assert page.locator(".icon-tabler-ad-off")

    events = []

    def cb(event):
        events.append(event)

    icon.param.watch(cb, "value")

    # test icon click updates value
    page.click(".bk-SVGIcon")
    wait_until(lambda: len(events) == 1, page)
    assert icon.value
    assert page.locator(".icon-tabler-ad-filled")


def test_toggle_icon_tabler_to_svg(page):
    tabler = "ad-off"

    icon = ToggleIcon(icon=tabler, active_icon=ACTIVE_SVG)
    serve_component(page, icon)

    # test defaults
    assert icon.icon == tabler
    assert not icon.value
    assert page.locator(".icon-tabler-ad-off")

    events = []

    def cb(event):
        events.append(event)

    icon.param.watch(cb, "value")

    # test icon click updates value
    page.click(".bk-TablerIcon")
    wait_until(lambda: len(events) == 1, page)
    assert icon.value
    assert page.locator(".icon-tabler-ad-filled")


def test_toggle_icon_svg_to_tabler(page):
    icon = ToggleIcon(icon=SVG, active_icon="ad-filled")
    serve_component(page, icon)

    # test defaults
    assert icon.icon == SVG
    assert not icon.value
    assert page.locator(".icon-tabler-ad-off")

    events = []

    def cb(event):
        events.append(event)

    icon.param.watch(cb, "value")

    # test icon click updates value
    page.click(".bk-SVGIcon")
    wait_until(lambda: len(events) == 1, page)
    assert icon.value
    assert page.locator(".icon-tabler-ad-filled")


def test_button_icon(page):
    icon = ButtonIcon(
        icon="clipboard", active_icon="check", toggle_duration=2000, size="5em"
    )
    serve_component(page, icon)

    # test defaults
    assert icon.icon == "clipboard"
    assert not icon.disabled
    assert page.locator(".clipboard")

    events = []

    def cb(event):
        events.append(event)

    icon.param.watch(cb, "clicks")

    # test icon click updates clicks
    page.click(".bk-TablerIcon")
    wait_until(lambda: len(events) == 1, page)
    assert len(events) == 1
    assert icon.clicks == 1
    assert page.locator(".check")

    # reverts back to original icon after toggle_duration
    wait_until(lambda: not icon.disabled, page)
    assert not icon.disabled
    assert page.locator(".clipboard")


def test_button_icon_disabled(page):
    icon = ButtonIcon(
        icon="clipboard",
        active_icon="check",
        toggle_duration=2000,
        size="5em",
        disabled=True,
    )
    serve_component(page, icon)

    # test defaults
    assert icon.icon == "clipboard"
    assert icon.disabled
    assert page.locator(".clipboard")

    events = []

    def cb(event):
        events.append(event)

    icon.param.watch(cb, "clicks")

    # test icon click does NOT update clicks
    page.click(".bk-TablerIcon")
    assert len(events) == 0
    assert icon.clicks == 0
    assert page.locator(".clipboard")

    # now enable the button
    icon.disabled = False
    wait_until(lambda: not icon.disabled, page)
    assert not icon.disabled
    assert page.locator(".clipboard")

    # test icon click updates clicks
    page.click(".bk-TablerIcon")
    wait_until(lambda: icon.disabled, page)
    wait_until(lambda: len(events) == 1, page)
    assert icon.clicks == 1
    assert page.locator(".check")

    # reverts back to original icon after toggle_duration
    wait_until(lambda: not icon.disabled, page, timeout=2100)
    assert page.locator(".clipboard")
    assert len(events) == 1


def test_button_icon_on_click_method(page):
    def on_click(event):
        static_text.value = f"Clicks: {button.clicks}"

    static_text = StaticText(value="Clicks: 0")
    button = ButtonIcon()
    button.on_click(on_click)
    row = Row(button, static_text)

    serve_component(page, row)

    page.click(".bk-TablerIcon")
    wait_until(lambda: static_text.value == "Clicks: 1", page)
    assert static_text.value == "Clicks: 1"

    button.clicks += 1
    assert static_text.value == "Clicks: 2"


def test_button_icon_on_click_kwarg(page):
    def on_click(event):
        static_text.value = f"Clicks: {button.clicks}"

    static_text = StaticText(value="Clicks: 0")
    button = ButtonIcon(on_click=on_click)
    row = Row(button, static_text)

    serve_component(page, row)

    page.click(".bk-TablerIcon")
    wait_until(lambda: static_text.value == "Clicks: 1", page)
    assert static_text.value == "Clicks: 1"

    button.clicks += 1
    assert static_text.value == "Clicks: 2"


def test_button_icon_js_on_click(page):
    button = ButtonIcon()
    int_slider = IntInput(value=0)
    button.js_on_click(args={"int_slider": int_slider}, code="int_slider.value += 1")
    row = Row(button, int_slider)

    serve_component(page, row)

    page.click(".bk-TablerIcon")
    wait_until(lambda: int_slider.value == 1, page)
    assert int_slider.value == 1


def test_button_icon_tooltip(page):
    button = ButtonIcon(description="Click me")
    serve_component(page, button)

    page.hover(".bk-TablerIcon")
    wait_until(lambda: page.locator(".bk-tooltip-content") is not None, page)
    assert page.locator(".bk-tooltip-content").text_content() == "Click me"


def test_button_icon_name(page):
    button = ButtonIcon(name="Like")
    serve_component(page, button)

    assert button.name == "Like"
    wait_until(lambda: page.locator(".bk-IconLabel").text_content() == "Like", page)


def test_button_icon_name_dynamically(page):
    button = ButtonIcon(name="Like")
    serve_component(page, button)

    assert button.name == "Like"
    assert page.locator(".bk-IconLabel").text_content() == "Like"

    button.name = "Dislike"
    assert button.name == "Dislike"
    wait_until(lambda: page.locator(".bk-IconLabel").text_content() == "Dislike", page)

    assert page.locator(".bk-IconLabel").bounding_box()["width"] < 40

    button.size = "2em"
    # check size
    wait_until(lambda: page.locator(".bk-IconLabel").bounding_box()["width"] >= 40, page)


def test_button_icon_description_dynamically(page):
    button = ButtonIcon(description="Like")
    serve_component(page, button)

    assert button.description == "Like"
    wait_until(lambda: page.locator(".bk-TablerIcon") is not None, page)
    page.locator(".bk-TablerIcon").click()
    assert page.locator(".bk-tooltip-content").text_content() == "Like"

    button.description = "Dislike"
    page.locator(".bk-TablerIcon").click()
    wait_until(lambda: page.locator(".bk-tooltip-content").text_content() == "Dislike", page)



import param

from panel.pane.placeholder import Placeholder
from panel.widgets import IntInput


class TestPlaceholder:

    def test_update_object(self):
        placeholder = Placeholder("Idle")
        placeholder.object = "Running..."
        assert placeholder.object == "Running..."
        placeholder.object = "New"
        assert placeholder.object == "New"

    def test_enter_exit(self):
        placeholder = Placeholder("⏳ Idle")
        with placeholder:
            placeholder.object = "🏃 Running..."
            assert placeholder.object == "🏃 Running..."
            placeholder.object = "🚶 Walking..."
            assert placeholder.object == "🚶 Walking..."
        assert placeholder.object == "⏳ Idle"

    def test_param_ref(self):
        int_input = IntInput(name="IntInput", start=1, end=10, value=5)
        placeholder = Placeholder(int_input.param.value)
        assert isinstance(placeholder.object, param.Integer)



import pytest

pytest.importorskip("playwright")

from playwright.sync_api import expect

from panel.pane import LaTeX
from panel.tests.util import serve_component

pytestmark = pytest.mark.ui

def test_latex_mathjax_renderer(page):
    ltx = LaTeX('$1+1$', renderer='mathjax')

    serve_component(page, ltx)

    expect(page.locator('mjx-container')).to_have_count(1)

def test_latex_katex_renderer(page):
    ltx = LaTeX('$1+1$', renderer='katex')

    serve_component(page, ltx)

    expect(page.locator('.katex-html')).to_have_count(1)
